import os
import subprocess
import re
import gzip
import sortedcontainers
import pandas as pd
import matplotlib
matplotlib.use('Agg')
from plotnine import *
import multiprocessing

def longReadMapping(strainName,longReads,reference,outputFolder,flag,longReadPlatform,threads):
    outputLog=""
    #Map with ngmlr
    samFile=os.path.join(outputFolder,strainName+".sam")
    p=subprocess.run(["ngmlr","-t",threads,"-x",longReadPlatform,"-r",reference,"-q",longReads,"-o",samFile],stderr=subprocess.PIPE,stdout=subprocess.PIPE, universal_newlines=True)
    outputLog+="COMMAND: "+" ".join(["ngmlr","-t",threads,"-x",longReadPlatform,"-r",reference,"-q",longReads,"-o",samFile])+"\n\n"
    if p.stderr!="" or p.stdout !="":
        outputLog+="STDERR:\n\n"+p.stderr+"\n\nSTDOUT:\n\n"+p.stdout+"\n\n"

    #Remove unmapped reads
    passSamFile=os.path.join(outputFolder,strainName+".pass.sam")
    p=subprocess.run(["samtools","view","-h","-t",reference,"-@",threads,"-F",flag,samFile,"-o",passSamFile],stderr=subprocess.PIPE,stdout=subprocess.PIPE, universal_newlines=True)
    outputLog+="COMMAND: "+" ".join(["samtools","view","-t",reference,"-@",threads,"-F",flag,samFile,"-o",passSamFile])+"\n\n"
    if p.stderr!="" or p.stdout !="":
        outputLog+="STDERR:\n\n"+p.stderr+"\n\nSTDOUT:\n\n"+p.stdout+"\n\n"

    #Sort sam file
    sortedHeaderSamFile=os.path.join(outputFolder,strainName+".sorted.header.sam")
    p=subprocess.run(["samtools","sort",passSamFile,"-@",threads,"-o",sortedHeaderSamFile],stderr=subprocess.PIPE,stdout=subprocess.PIPE, universal_newlines=True)
    outputLog+="COMMAND: "+" ".join(["samtools","sort",passSamFile,"-@",threads,"-o",sortedHeaderSamFile])+"\n\n"
    if p.stderr!="" or p.stdout !="":
        outputLog+="STDERR:\n\n"+p.stderr+"\n\nSTDOUT:\n\n"+p.stdout+"\n\n"

    #This looks stupid but is the safest way I could think of to get rid of the SAM header (it gets in the way of another function later)
    sortedSamFile=os.path.join(outputFolder,strainName+".sorted.sam")
    p=subprocess.run(["samtools","view",sortedHeaderSamFile,"-@",threads,"-o",sortedSamFile],stderr=subprocess.PIPE,stdout=subprocess.PIPE, universal_newlines=True)
    outputLog+="COMMAND: "+" ".join(["samtools","view",sortedHeaderSamFile,"-@",threads,"-o",sortedSamFile])+"\n\n"
    if p.stderr!="" or p.stdout !="":
        outputLog+="STDERR:\n\n"+p.stderr+"\n\nSTDOUT:\n\n"+p.stdout+"\n\n"

    try:
        #Cleaning up
        os.remove(samFile)
        os.remove(passSamFile)
        os.remove(sortedHeaderSamFile)
    except:
        return outputLog,"Long reads not mapped to reference"

    return outputLog,"Long reads successfully mapped to reference"

def shortReadMapping(strainName,R1,R2,reference,outputFolder,threads):
    outputLog=""

    #Mapping
    samFileNoRGPath=os.path.join(outputFolder, strainName+".noRG.sam")
    samFileNoRG=open(samFileNoRGPath,"w")
    p=subprocess.run(["bwa","mem","-t",threads,"-M", reference, R1, R2],stderr=subprocess.PIPE,stdout=samFileNoRG, universal_newlines=True)
    samFileNoRG.close()
    outputLog+="COMMAND: "+" ".join(["bwa","mem","-t",threads,"-M", reference, R1, R2, ">", samFileNoRGPath])+"\n\n"
    if p.stderr!="":
        outputLog+="STDERR:\n\n"+p.stderr+"\n\nSTDOUT is in "+samFileNoRGPath+"\n\n"

    #Adding read groups
    samFile=os.path.join(outputFolder, strainName+".sam")
    p=subprocess.run(["gatk", "AddOrReplaceReadGroups","-I",samFileNoRGPath,"-O",samFile,"-RGID","ID_"+strainName, "-RGLB","LB_"+strainName, "-RGPL","ILLUMINA", "-RGPU","PU_"+strainName, "-RGSM","SM_"+strainName],stderr=subprocess.PIPE,stdout=subprocess.PIPE, universal_newlines=True)
    outputLog+="COMMAND: "+" ".join(["gatk", "AddOrReplaceReadGroups","-I",samFileNoRGPath,"-O",samFile,"-RGID","ID_"+strainName, "-RGLB","LB_"+strainName, "-RGPL","ILLUMINA", "-RGPU","PU_"+strainName, "-RGSM","SM_"+strainName])+"\n\n"
    if p.stderr!="" or p.stdout!="":
        outputLog+="STDERR:\n\n"+p.stderr+"\n\nSTDOUT:\n\n"+p.stdout+"\n\n"

    #Organizing mapped reads
    bamFile=os.path.join(outputFolder,strainName+".bam")
    sortedBamFile=os.path.join(outputFolder,strainName+".sorted.bam")

    print("Removing quality 0 (multimapped) reads, turning to bam and sorting it")

    p=subprocess.run(["samtools","view","-@",threads,"-bT",reference,"-q","1",samFile,"-o",bamFile],stderr=subprocess.PIPE, stdout=subprocess.PIPE, universal_newlines=True)
    outputLog+="COMMAND: "+" ".join(["samtools","view","-@",threads,"-bT",reference,"-q","1",samFile,"-o",bamFile])+"\n\n"
    if p.stderr!="" or p.stdout !="":
        outputLog+="STDERR:\n\n"+p.stderr+"\n\nSTDOUT:\n\n"+p.stdout+"\n\n"

    p=subprocess.run(["samtools", "sort", bamFile,"-@",threads, "-o", sortedBamFile],stderr=subprocess.PIPE, stdout=subprocess.PIPE, universal_newlines=True)
    outputLog+="COMMAND: "+" ".join(["samtools", "sort", bamFile,"-@",threads, "-o", sortedBamFile])+"\n\n"
    if p.stderr!="" or p.stdout !="":
        outputLog+="STDERR:\n\n"+p.stderr+"\n\nSTDOUT:\n\n"+p.stdout+"\n\n"

    p=subprocess.run(["samtools", "index","-@",threads, sortedBamFile],stderr=subprocess.PIPE, stdout=subprocess.PIPE, universal_newlines=True)
    outputLog+="COMMAND: "+" ".join(["samtools", "index","-@",threads, sortedBamFile])+"\n\n"
    if p.stderr!="" or p.stdout !="":
        outputLog+="STDERR:\n\n"+p.stderr+"\n\nSTDOUT:\n\n"+p.stdout+"\n\n"

    #GATK cleaning
    MDsortedBamFile=os.path.join(outputFolder,strainName+".sorted.MD.bam")
    p=subprocess.run(["gatk","MarkDuplicates", "--REMOVE_DUPLICATES", "true", "-I",sortedBamFile,"-O",MDsortedBamFile,"-M", "/dev/null"],stderr=subprocess.PIPE, stdout=subprocess.PIPE, universal_newlines=True)
    outputLog+="COMMAND: "+" ".join(["gatk","MarkDuplicates", "--REMOVE_DUPLICATES", "true", "-I",sortedBamFile,"-O",MDsortedBamFile,"-M", "/dev/null"])+"\n\n"
    if p.stderr!="" or p.stdout !="":
        outputLog+="STDERR:\n\n"+p.stderr+"\n\nSTDOUT:\n\n"+p.stdout+"\n\n"

    #Finalizing
    finalBamFile=os.path.join(outputFolder,strainName+".final.bam")
    p=subprocess.run(["samtools", "sort", MDsortedBamFile, "-@",threads,"-o", finalBamFile],stderr=subprocess.PIPE, stdout=subprocess.PIPE, universal_newlines=True)
    outputLog+="COMMAND: "+" ".join(["samtools", "sort", MDsortedBamFile, "-@",threads,"-o", finalBamFile])+"\n\n"
    if p.stderr!="" or p.stdout !="":
        outputLog+="STDERR:\n\n"+p.stderr+"\n\nSTDOUT:\n\n"+p.stdout+"\n\n"

    p=subprocess.run(["samtools", "index","-@",threads, finalBamFile],stderr=subprocess.PIPE, stdout=subprocess.PIPE, universal_newlines=True)
    outputLog+="COMMAND: "+" ".join(["samtools", "index","-@",threads, finalBamFile])+"\n\n"
    if p.stderr!="" or p.stdout !="":
        outputLog+="STDERR:\n\n"+p.stderr+"\n\nSTDOUT:\n\n"+p.stdout+"\n\n"

    try:
        #Cleaning up
        os.remove(samFileNoRGPath)
        os.remove(samFile)
        os.remove(bamFile)
        os.remove(sortedBamFile)
        os.remove(MDsortedBamFile)
    except:
        return outputLog,"ERROR: Short reads unsuccessfully mapped to reference"

    return outputLog,"Short reads successfully mapped to reference"

def getShortReadPositions(hetVCFName,shortReadPosFileName):
    hetVCFFile=open(hetVCFName,"r")

    hetVCFPositions=set()

    for line in hetVCFFile:
        line=line.strip("\n").split("\t")
        if "#" not in line[0]:
            chr=line[0].replace(":","_")
            pos=line[1]
            ref=line[3]
            alts=line[4].split(",")
            for alt in alts:
                if len(alt)==1 and len(ref)==1: #SNP or del of 1 base ("*")
                    hetVCFPositions.add(chr+":"+pos)
                elif len(alt)>len(ref): #INS
                    #hetVCFPositions.add(chr+":"+pos)
                    pass
                elif len(alt)<len(ref): #DEL
                    #for curPos in range(int(line[1])+1,int(line[1])+len(ref)):
                    #   hetVCFPositions.add(chr+":"+str(curPos))
                    pass

    #Writing the short read het positions to a file
    shortReadSNPText=""

    for position in hetVCFPositions:
        shortReadSNPText+=position+"\n"

    shortReadSNPFile=open(shortReadPosFileName,"w")

    shortReadSNPFile.write(shortReadSNPText)

    shortReadSNPFile.close()

    return "Determined positions of heterozygous SNPs based on short read data"

def assignLongReadToSNPs(samPath,bedPath,referencePath,minQ,minMQ,minAln,outputPath):
    ###########
    #Resources#
    ###########

    #Cigar string:
    #http://bioinformatics.cvr.ac.uk/blog/tag/cigar-string/
    #http://zenfractal.com/2013/06/19/playing-with-matches/
    #https://davetang.org/wiki/tiki-index.php?page=SAM

    #QScores
    #https://medium.com/@robertopreste/phred-quality-score-2837415f0af

    #To consider:
    #Currently, softclips, insertions and padding worsen the alignment score, you might want to change that

    cigarPartRE='(\d+)([DHIMNPSX=])'

    #Initializing QScores

    #P = 10-Q/10

    QScores={}
    for i in range(0,200):
        QScores[chr(33+i)]=i

    PScores={}

    for ASCII, Q in QScores.items():
        PScores[ASCII]=10**(-Q/10)

    #Here it takes the different bedlines and it tries to apply the 0-base 1-base conversion maybe? #We'll only print things in these intervals
    variantSet=set()

    bedFile=open(bedPath,"r")

    for line in bedFile:
        line=line.strip("\n").split("\t")
        contig=line[0].replace(":","_")
        start=str(int(line[1])+1) #The bedFile is in 0-base, we want to stay in 1-base
        variantSet.add(contig+":"+start)

    bedFile.close()

    #This takes all the samfile lines
    samLines=[]

    samFile=open(samPath,"r")
    for line in samFile:
        line=line.strip("\n").split("\t")
        if line[0][0]!="@":
            line[2]=line[2].replace(":","_")
            samLines.append(line)

    samFile.close()

    manager = multiprocessing.Manager()

    readDict = manager.dict()

    with multiprocessing.Pool(processes=8) as pool: #Consider batching lines
        pool.map(assignLongReadToSNPsParallelHelper, [[line,readDict,QScores,variantSet] for line in samLines])

    outputText=""

    for readName, values in readDict.items():
        lineContents=readName+"\t"+"\t".join(values)+"\n"
        outputText+=lineContents

    outputFile=open(outputPath,"w")
    outputFile.write(outputText)
    outputFile.close()

    return "Reduced long reads to heterozygous SNPs based on short read information"

def assignLongReadToSNPsParallelHelper(variables):
    line=variables[0]
    rD=variables[1]
    QScores=variables[2]
    variantSet=variables[3]
    assignLongReadToSNPsParallel(line,rD,QScores,variantSet)

def assignLongReadToSNPsParallel(line,readDict,QScores,variantSet):
    cigarPartRE = '(\d+)([DHIMNPSX=])'
    if line[0] not in readDict:
        readDict[line[0]]=[]
    else:
        i=1
        b=False
        while b==False:
            if line[0]+"_"+str(i) not in readDict:
                readDict[line[0]+"_VCSTTP_"+str(i)]=[] #Very Clean Solution To This Problem
                line[0]=line[0]+"_VCSTTP_"+str(i)
                b=True
            i+=1

    #Initializing a lot of things
    sequence=line[9]
    if sequence=="*": #Only seen this happen when the MAPQ is 0 for a secondary alignment
        return
    readName=line[0]
    ctgName=line[2]
    refPosition=int(line[3]) #SamFiles are already in 1-base
    MQ=int(line[4])
    cigar=line[5]
    QLine=line[10]
    readPosition=0 #The base of readPosition is irrelevant, it starts at 0 because python lists are 0-based (sequence[0])
    lineBases=[]
    matchStart=False
    averageScore=0
    for Q in QLine:
        try:
            averageScore+=QScores[Q]
        except:
            QScores[Q]=ord(Q)-33
            averageScore+=QScores[Q]
    averageScore=int(averageScore/len(QLine))
    #
    #Okay so now we evaluate the cigar string and just slide down the SEQ and the reference simultaneously
    for match in re.finditer(cigarPartRE, cigar):
        n=int(match.group(1))
        alignmentType=match.group(2)
        if alignmentType=="S":
            readPosition+=n
        elif alignmentType in ["M", "=", "X"]:
            matchStart=True
            matches=[] #So, we know there might be some mismatches in here
            for x in range(n):
                matches.append((refPosition,sequence[readPosition]))
                refPosition+=1
                readPosition+=1
            for match in matches:
                pos=match[0]
                base=match[1]
                posName=ctgName+":"+str(pos)
                if posName in variantSet:
                    lineBases.append(posName+"="+base) #+"_"+str(QScores[Q])+"-"+str(averageScore)) #For debugging/info about which parameters can predict good reads
        elif alignmentType=="I":
            insertionSequence=""
            for x in range(n):
                insertionSequence+=sequence[readPosition]
                readPosition+=1
            pos=refPosition
            posName=ctgName+":"+str(pos)
            if posName in variantSet:
                #lineBases.append(posName+"="+insertionSequence) #+"_"+str(insertionScore)+"-"+str(averageScore)) #For debugging/info about which parameters can predict good reads
                pass #WE IGNORE THESE FOR NOW
        elif alignmentType=="D":
            if matchStart:
                for x in range(n):
                    matches.append((refPosition,"D",averageScore,averageScore))
                    refPosition+=1
                for match in matches:
                    pos=match[0]
                    base=match[1] #This will automatically be "D" for deletion
                    posName=ctgName+":"+str(pos)
                    if posName in variantSet:
                        #lineBases.append(posName+"="+base) #+"_"+str(Q)+"-"+str(averageScore)) #For debugging/info about which parameters can predict good reads
                        pass #WE IGNORE THESE FOR NOW
    readDict[readName]=lineBases

def longReadValidation(longReadFilePath,minCov,minRatio,minTrioCov,validatedSNPAssignmentsFile,contextDepthsFile):
    longReadFile=open(longReadFilePath,"r")

    longReads=[]

    for read in longReadFile:
        longReads.append(read.strip("\n").split("\t"))

    longReadFile.close()

    contextDepths={}
    batchSize=3000
    batches=[]
    currentBatch=[]
    i=0
    for longRead in longReads:
        currentBatch.append(longRead)
        i+=1
        if i>batchSize:
            batches.append(currentBatch)
            currentBatch=[]
            i=0

    if currentBatch!=[]:
        batches.append(currentBatch)

    with multiprocessing.Pool(processes=8) as pool:
        contextDepthDicts=pool.map(updateContextDepths, [batchData for batchData in batches])
        for contextDepthDict in contextDepthDicts:
            for SNP, SNPContexts in contextDepthDict.items():
                contextDepths.setdefault(SNP,{})
                for SNPContext, SNPContextCount in SNPContexts.items():
                    contextDepths[SNP].setdefault(SNPContext,0)
                    contextDepths[SNP][SNPContext]+=SNPContextCount

    contextText=""

    for SNP, contexts in contextDepths.items():
        contextTextLine=SNP+"\t"
        for contextSNPs, quantity in contexts.items():
            contextTextLine+=" ".join(contextSNPs)+"\t"+str(quantity)+"\t"
        contextTextLine=contextTextLine.strip("\t")
        contextText+=contextTextLine+"\n"

    contextFile=open(contextDepthsFile,"w")
    contextFile.write(contextText)
    contextFile.close()


    longReadSets={}
    longReadTuples={}


    ###########################################################################################
    #You can't just do this step before you remove reads based on coverage, that's ridiculous.#
    #vvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvvv#
    ###########################################################################################

    #Only keeping long reads that have unique combinations of SNPs
    uniqueLongReadSets=set()
    uniqueLongReadIDs=set()

    curLen=len(uniqueLongReadSets)

    for read in longReads:
        readSet=frozenset(read[1:])
        altruisticSet=[]
        for SNP in readSet:
            altruisticSet.append(SNP)
        readSet=frozenset(altruisticSet)
        uniqueLongReadSets.add(readSet)
        if len(uniqueLongReadSets)>curLen:
            curLen+=1
            longReadSets[read[0]]=readSet
            longReadTuples[read[0]]=tuple(readSet)
            uniqueLongReadIDs.add(read[0])

    ###########################################################################################
    #^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^^#
    #You can't just do this step before you remove reads based on coverage, that's ridiculous.#
    ###########################################################################################

    SNPAssignments=longReadSets

    #Printing it all to a file
    SNPAssignmentsText=""

    for longRead, assignedSNPs in SNPAssignments.items():
        if assignedSNPs!=set({''}):
            SNPAssignmentsText+=longRead+"\t"+"\t".join(assignedSNPs)+"\n"

    SNPAssignmentFile=open(validatedSNPAssignmentsFile,"w")
    SNPAssignmentFile.write(SNPAssignmentsText)
    SNPAssignmentFile.close()

    return "Pre-processed long reads and generated context depth file"

def updateContextDepths(longReadBatch):
    localContextDepths={}
    distance=2
    for longRead in longReadBatch:
        SNPs=longRead[1:]
        for SNP in SNPs:
            i=SNPs.index(SNP)
            itemList=[]
            for j in range(i-distance,i+distance+1):
                if j>=0 and j<len(SNPs):
                    itemList.append(SNPs[j])
            if SNP!='':
                frozenList=frozenset(itemList)
                localContextDepths.setdefault(SNP,{})
                localContextDepths[SNP].setdefault(frozenList,0)
                localContextDepths[SNP][frozenList]+=1
    return localContextDepths

def getAvgFastQScore(fastQScoreStr):
    fastQScoreStr=fastQScoreStr.strip()
    totalScore=0
    QScores={}
    for baseQ in fastQScoreStr:
        if baseQ not in QScores:
            QScores[baseQ]=ord(baseQ)-33
        totalScore+=QScores[baseQ]
    avgScore=totalScore/len(fastQScoreStr)
    return avgScore

def generateLongReadFastQFiles(haplotigReadNameFilePath,longReadFastQFilePath,outputPath): #Parallelize in a way that can save memory
    ##clusterStatsText=""
    haplotigReadDict={}
    allAllowedReads=set()
    haplotigReadNameFile=open(haplotigReadNameFilePath,"r")
    for line in haplotigReadNameFile:
        line=line.strip("\n").split("\t")
        haplotigName=line[0]
        readName="@"+line[1]
        readName=readName.split("_VCSTTP_")[0]
        haplotigReadDict.setdefault(readName,set())
        haplotigReadDict[readName].add(haplotigName)
        allAllowedReads.add(readName)

    haplotigReadNameFile.close()


    ##longReadMetaData={}

    try:
        longReadFastQFile=gzip.open(longReadFastQFilePath,"rt")
        longReadFastQFile.readline()
        longReadFastQFile.close()
        longReadFastQFile=gzip.open(longReadFastQFilePath, "rt")
    except gzip.BadGzipFile:
        longReadFastQFile=open(longReadFastQFilePath,"r")

    fastQReadQueue = multiprocessing.Queue()

    fastQWriter_p = multiprocessing.Process(target=fastQWriter, args=((outputPath),(fastQReadQueue),))
    fastQWriter_p.daemon = True
    fastQWriter_p.start()

    i=0
    j=0
    readData=[]
    readDataBatch={}
    batchSize=2500
    for line in longReadFastQFile:
        line=line.strip("\n")
        if i%4==0:
            if readData!=[]:
                for haplotigName in haplotigNames:
                    readDataBatch.setdefault(haplotigName,[])
                    readDataBatch[haplotigName].append(readData)
                    j+=1
                if j>batchSize:
                    j=0
                    fastQReadQueue.put(readDataBatch)
                    readDataBatch={}
            line=line.split()
            readName=line[0]
            haplotigNames=[]
            readData=[]
            if readName in allAllowedReads:
                for haplotigName in haplotigReadDict[readName]:
                    haplotigNames.append(haplotigName)
                readData=[readName]
        else:
            if readName in allAllowedReads:
                ##if i%4==3:
                    ##longReadMetaData[readName]=getAvgFastQScore(line)
                readData.append(line)
        i+=1
    longReadFastQFile.close()

    if readDataBatch!={}:
        fastQReadQueue.put(readDataBatch)

    fastQReadQueue.put("end")

    fastQWriter_p.join()

    ##clusterStatsFile=open(outputPath+"clusterStats.tsv", 'w')
    ##clusterStatsFile.write(clusterStatsText)
    ##clusterStatsFile.close()

    return "Successfully generated phased FastQ files"

def fastQWriter(outputPath,queue):
    while True:
        readDataBatch=queue.get()
        if readDataBatch == "end":
            return
        for haplotigName, haplotigReads in readDataBatch.items():
            haplotigFastQText=""
            for readData in haplotigReads:
                readName=readData[0]
                readString=readName+"\n"+"\n".join(readData[1:])+"\n"
                haplotigFastQText+=readString
            haplotigFile=gzip.open(outputPath+haplotigName+".fastq.gz",'at')
            haplotigFile.write(haplotigFastQText)
            haplotigFile.close()

def loadFile(filePath):
        openFile=open(filePath,"r")
        fileContents=[]
        for line in openFile:
                line=line.strip("\n").split("\t")
                fileContents.append(line)
        openFile.close()
        return fileContents

def simplifyDataVis(dataVisPath,simpleOutPath,distance):
    fileContents=loadFile(dataVisPath)
    fileDict={}

    for line in fileContents:
        if line[0] not in fileDict:
            fileDict[line[0]]=[line]
        else:
            fileDict[line[0]].append(line)

    newLines=[]

    for haplotig, SNPLines in fileDict.items():
        startPos=int(SNPLines[0][1])
        endPos=int(SNPLines[0][1])+1
        for line in SNPLines:
            if int(line[1])>endPos+distance:
                newLines.append([haplotig,str(startPos),str(endPos),line[2],line[3]])
                startPos=int(line[1])
                endPos=int(line[1])+1
            else:
                endPos=int(line[1])
        newLines.append([haplotig, str(startPos), str(endPos), line[2], line[3]])

    outFileText=""
    for line in newLines:
        outFileText+="\t".join(line)+"\n"

    outFile=open(simpleOutPath,"w")
    outFile.write(outFileText)
    outFile.close()

def generatePhasingVis(dataVisPath,outPath):
    #Figure out how supported (as a %) each base is for each position in the cluster

    outputPNG=outPath+"phasedVis.png"
    outputPDF=outPath+"phasedVis.pdf"
    outputSVG=outPath+"phasedVis.svg"

    tbl=pd.read_csv(dataVisPath,sep="\t",header=None)
    tbl.columns=["contigName","startPos","endPos","chr","yValue"]

    g=(ggplot(tbl,aes(y="contigName",yend="contigName",x="startPos",xend="endPos",color='contigName'))+geom_segment(size=1.5)+theme(legend_position="none")+theme(panel_grid_minor=element_blank())+facet_wrap("~chr",scales="free")+theme(axis_title_y=element_blank(),axis_text_y=element_blank(),axis_ticks_major_y=element_blank())+xlab("Position (bp)")+theme(subplots_adjust={'hspace':0.7}))

    ggsave(g,filename=outputSVG,width=18,height=10)
    ggsave(g,filename=outputPNG,width=18,height=10)
    ggsave(g,filename=outputPDF,width=18,height=10)

    return "Generated phased plots"

def generateCoverageVis(dataVisPath,outPath):
    #Figure out how covered (as X) each heterozygous position is for each cluster

    outputPNG=outPath+"coverageVis.png"
    outputPDF=outPath+"coverageVis.pdf"
    outputSVG=outPath+"coverageVis.svg"

    tbl=pd.read_csv(dataVisPath,sep="\t",header=None,na_values="NA")
    tbl.columns=["haplotigName","chr","pos","coverage"]

    g=(ggplot(tbl,aes(x="pos",y="coverage",color='haplotigName'))+geom_line(size=1.5)+theme(legend_position="none")+theme(panel_grid_minor=element_blank())+facet_wrap("~chr",scales="free")+ylab("Coverage (X)")+xlab("Position (bp)")+theme(subplots_adjust={'hspace': 0.7}))

    ggsave(g,filename=outputSVG,width=18,height=10)
    ggsave(g,filename=outputPNG,width=18,height=10)
    ggsave(g,filename=outputPDF,width=18,height=10)

    return "Generated coverage plots"

def generateDiscordanceVis(dataVisPath,outPath): #This function is very slow with plotnine for some reason. Should replace this in the long run.
    #Figure out how supported (as a %) each base is for each position in the cluster

    #Takes several minutes to output each graph and violin plots have an unreasonable memory constraint using plotnine.
    #TODO: Fix this
    #TEMP: Moved this to very last part of the pipeline so that everything else can be generated smoothly first.
    #TEMP: Commented out two of the less useful file formats to save time.

    #outputPNG=outPath+"discordanceVis.png"
    #outputPDF=outPath+"discordanceVis.pdf"
    outputSVG=outPath+"discordanceVis.svg"

    tbl=pd.read_csv(dataVisPath,sep="\t",header=None,comment="#")
    tbl.columns=["cluster","chr","position","base","frequency","coverage"]

    g=(ggplot(tbl,aes(y="frequency",x='cluster'))+facet_wrap("~chr",scales="free")+geom_violin(scale="width")+coord_flip()+ylim(0,1)+xlab("Cluster name")+ylab("Allele frequency (%)")+ggtitle("Allele frequency by cluster (all clusters have same max width)")+theme(subplots_adjust={'wspace': 0.25}))

    #ggsave(g,filename=outputPNG,width=18,height=10)
    #ggsave(g,filename=outputPDF,width=18,height=10)
    ggsave(g,filename=outputSVG,width=18,height=10)

    return "Generated discordance plots"

def generateDiscordance(clusterReadFilePath,readFilePath,outPath):
    #Figure out how supported (as a %) each base is for each position in the cluster

    #Identify warnings/sort by how bad & output a text file

    #Output a plot
    clusterDict={} #All reads associated with clusters

    clusterFile=open(clusterReadFilePath,"r")

    for line in clusterFile:
        line=line.strip("\n").split("\t")
        if line[0] not in clusterDict:
            clusterDict[line[0]]=set()
        clusterDict[line[0]].add(line[1])

    clusterFile.close()

    readDict={} #All SNPs associated with reads

    readFile=open(readFilePath,"r")

    for line in readFile:
        line=line.strip("\n").split("\t")
        if line[0] not in readDict:
            readDict[line[0]]=set()
        readDict[line[0]].update(line[1:])

    readFile.close()

    rawDict={} #All SNPs within clusters based on the reads involved

    for cluster,reads in clusterDict.items():
        rawDict[cluster]={}
        for read in reads:
            for SNP in readDict[read]:
                chr=SNP.split(":")[0]
                position=SNP.split(":")[1].split("=")[0]
                base=SNP.split("=")[1]
                if chr not in rawDict[cluster]:
                    rawDict[cluster][chr]={}
                if position not in rawDict[cluster][chr]:
                    rawDict[cluster][chr][position]={}
                if base not in rawDict[cluster][chr][position]:
                    rawDict[cluster][chr][position][base]=0
                rawDict[cluster][chr][position][base]+=1 #Count the number for each base

    allFrequencies=[]

    for cluster,chrs in rawDict.items():
        for chr,positions in chrs.items():
            for position, bases in positions.items():
                baseCovs=bases.values()
                totalCov=sum(baseCovs)
                for base, coverage in bases.items():
                    allFrequencies.append((cluster,chr,position,base,str(coverage/totalCov),str(totalCov)))

    cleanFullText="#cluster\tchr\tposition\tbase\tfrequency\tcoverage\n"

    for freq in allFrequencies:
        cleanFullText+="\t".join(freq)+"\n"

    cleanFullTextOutputFile=open(outPath,"w")
    cleanFullTextOutputFile.write(cleanFullText)
    cleanFullTextOutputFile.close()

    pass

def generateCoverage(clusterFilePath,readFilePath,outPath,windowSize):
    #Figure out how covered each read position is in each cluster

    totalCoverage=0
    uniquePositions=set()

    clusterDict={}

    clusterFile=open(clusterFilePath,"r")

    for line in clusterFile:
        line=line.strip("\n").split("\t")
        if line[0] not in clusterDict:
            clusterDict[line[0]]=set()
        clusterDict[line[0]].add(line[1])

    clusterFile.close()

    readDict={}

    readFile=open(readFilePath,"r")

    for line in readFile:
        line=line.strip("\n").split("\t")
        if line[0] not in readDict:
            readDict[line[0]]=set()
        readDict[line[0]].update(line[1:])

    readFile.close()

    rawDict={}

    for cluster,reads in clusterDict.items():
        rawDict[cluster]={}
        for read in reads:
            for SNP in readDict[read]:
                uniquePositions.add(SNP.split("=")[0])
                chr=SNP.split(":")[0]
                position=SNP.split(":")[1].split("=")[0]
                if chr not in rawDict[cluster]:
                    rawDict[cluster][chr]={}
                if position not in rawDict[cluster][chr]:
                    rawDict[cluster][chr][position]=0
                rawDict[cluster][chr][position]+=1
                totalCoverage+=1

    cleanFullText=""

    betterClusters={}

    for cluster,chrs in rawDict.items():
        betterClusters[cluster]={}
        for chr,positions in rawDict[cluster].items():
            betterClusters[cluster][chr]=[]
            for pos in positions:
                cov=str(rawDict[cluster][chr][pos])
                if chr not in betterClusters[cluster].keys():
                    betterClusters[cluster][chr]=[]
                betterClusters[cluster][chr].append((int(pos),int(cov)))
            sortedPositions=betterClusters[cluster][chr]
            sortedPositions.sort()
            betterClusters[cluster][chr]=sortedPositions

    windowDict={}

    for cluster,chrs in betterClusters.items():
        windowDict[cluster]={}
        for chr in chrs.keys():
            if chr not in windowDict[cluster]:
                windowDict[cluster][chr]={}
            sortedPositions=betterClusters[cluster][chr]
            sortedPositions.sort()
            minPos=0
            maxPos=int(sortedPositions[-1][0])+windowSize
            for i in range(minPos,maxPos,windowSize):
                window=[]
                for position in sortedPositions:
                    if int(position[0])>=i and int(position[0])<i+windowSize:
                        window.append(int(position[1]))
                if len(window)==0:
                    mean="NA"
                else:
                    mean=sum(window)/len(window)
                    if mean==0:
                        mean="NA"
                windowDict[cluster][chr][i+(windowSize/2)]=str(mean)

    cleanFullText=""

    for cluster,chrs in windowDict.items():
        for chr,positions in chrs.items():
            for pos,cov in positions.items():
                cleanFullText+=cluster+"\t"+chr+"\t"+str(pos)+"\t"+str(cov)+"\n"

    cleanFullTextOutputFile=open(outPath,"w")
    cleanFullTextOutputFile.write(cleanFullText)
    cleanFullTextOutputFile.close()

    minCov=int(0.05*(totalCoverage/len(uniquePositions)))

    return minCov


def giveMeFullData(clusters):
    clusterText=""
    sortedClusterLines=sortedcontainers.SortedList()
    clusterLines=[]
    for clusterName, cluster in clusters.items():
        for SNP in cluster:
            contig=SNP.split(":")[0]
            position=int(SNP.split(":")[1].split("=")[0])
            sortedClusterLines.add([position,clusterName,contig])
    i=1
    previouslySeen={sortedClusterLines[0][1]:i}
    for line in sortedClusterLines:
        if line[1] not in previouslySeen:
            i+=1
            previouslySeen[line[1]]=i
        clusterLines.append([line[1],line[0],line[2],previouslySeen[line[1]]])
    for line in clusterLines:
        clusterText+="\t".join([str(x) for x in line])+"\n"
    return clusterText

def filterCoverage(clusterReadFilePath,readFilePath,outPath,minCov):
    #Filter out how based on coverage

    clusterDict={} #All reads associated with clusters

    clusterFile=open(clusterReadFilePath,"r")

    for line in clusterFile:
        line=line.strip("\n").split("\t")
        if line[0] not in clusterDict:
            clusterDict[line[0]]=set()
        clusterDict[line[0]].add(line[1])

    clusterFile.close()

    readDict={} #All SNPs associated with reads

    readFile=open(readFilePath,"r")

    for line in readFile:
        line=line.strip("\n").split("\t")
        if line[0] not in readDict:
            readDict[line[0]]=set()
        readDict[line[0]].update(line[1:])

    readFile.close()

    rawDict={} #All SNPs within clusters based on the reads involved

    for cluster,reads in clusterDict.items():
        rawDict[cluster]={}
        for read in reads:
            for SNP in readDict[read]:
                chr=SNP.split(":")[0]
                position=SNP.split(":")[1].split("=")[0]
                base=SNP.split("=")[1]
                if chr not in rawDict[cluster]:
                    rawDict[cluster][chr]={}
                if position not in rawDict[cluster][chr]:
                    rawDict[cluster][chr][position]=0
                rawDict[cluster][chr][position]+=1 #Count the number for each base

    allFrequencies=[]

    for cluster,chrs in rawDict.items():
        for chr,positions in chrs.items():
            for position, coverage in positions.items():
                if coverage>minCov:
                    allFrequencies.append((cluster,chr,position,str(coverage)))

    cleanFullText="#cluster\tchr\tcoverage\tposition\n"

    for freq in allFrequencies:
        cleanFullText+="\t".join(freq)+"\n"

    cleanFullTextOutputFile=open(outPath,"w")
    cleanFullTextOutputFile.write(cleanFullText)
    cleanFullTextOutputFile.close()

    pass








import argparse
import sys
import os
import subprocess
import bin.nPhasePipelineFunctions as nPhaseFunctions
import bin.nPhase as phaseTool
import bin.nPhaseCleaning as cleaningTool

def updateLog(logFilePath,logText):
    logFile=open(logFilePath,"a")
    logFile.write(logText)
    logFile.close()


def nPhasePipeline(args):
    #Create folder structure
    basePath=os.path.join(args.outputFolder,args.strainName)
    mappedShortReadPath=os.path.join(basePath,"Mapped","shortReads")
    variantCalledShortReadPath=os.path.join(basePath,"VariantCalls","shortReads")
    mappedLongReadPath=os.path.join(basePath,"Mapped","longReads")
    variantCalledLongReadPath=os.path.join(basePath,"VariantCalls","longReads")
    overlapPath=os.path.join(basePath,"Overlaps")
    phasedPath=os.path.join(basePath,"Phased")
    phasedFastqPath=os.path.join(basePath,"Phased","FastQ")
    datavisFolderPath=os.path.join(basePath,"Phased","Plots")
    logPath=os.path.join(basePath,"Logs") #Currently not producing any logs

    allPaths=[basePath,mappedShortReadPath,variantCalledShortReadPath,mappedLongReadPath,variantCalledLongReadPath,overlapPath,phasedPath,phasedFastqPath,datavisFolderPath,logPath]

    for path in allPaths:
        os.makedirs(path, exist_ok=True)

    readmePath=os.path.join(basePath,"Readme.txt")
    fullLogPath=os.path.join(basePath,"Logs","fullLog.txt")

    ########################
    #Pre-process reference #
    ########################

    #Make sure the reference is indexed (might be missing a process)
    if not os.path.isfile(args.reference+".fai"):
        p=subprocess.run(["samtools","faidx",args.reference],stderr=subprocess.PIPE,stdout=subprocess.PIPE, universal_newlines=True)
        logText="COMMAND: "+" ".join(["samtools","faidx",args.reference])+"\n\n"
        if p.stderr!="" or p.stdout !="":
            logText+="STDERR:\n\n"+p.stderr+"\n\nSTDOUT:\n\n"+p.stdout+"\n\n"
        updateLog(fullLogPath,logText)

    if not os.path.isfile(args.reference+".bwt"):
        p=subprocess.run(["bwa","index",args.reference],stderr=subprocess.PIPE,stdout=subprocess.PIPE, universal_newlines=True)
        logText="COMMAND: "+" ".join(["bwa","index",args.reference])+"\n\n"
        if p.stderr!="" or p.stdout !="":
            logText+="STDERR:\n\n"+p.stderr+"\n\nSTDOUT:\n\n"+p.stdout+"\n\n"
        updateLog(fullLogPath,logText)

    if not os.path.isfile(args.reference+".dict"):
        p=subprocess.run(["gatk","CreateSequenceDictionary","-R",args.reference],stderr=subprocess.PIPE,stdout=subprocess.PIPE, universal_newlines=True)
        logText="COMMAND: "+" ".join(["gatk","CreateSequenceDictionary","-R",args.reference])+"\n\n"
        if p.stderr!="" or p.stdout !="":
            logText+="STDERR:\n\n"+p.stderr+"\n\nSTDOUT:\n\n"+p.stdout+"\n\n"
        updateLog(fullLogPath,logText)

    ########################
    #Pre-process long reads#
    ########################

    #Map long reads to reference
    splitReadFlag="260" #This flag allows split reads
    outputLog,systemMessage=nPhaseFunctions.longReadMapping(args.strainName,args.longReadFile,args.reference,mappedLongReadPath,splitReadFlag,args.longReadPlatform,args.threads)
    print(systemMessage)
    updateLog(fullLogPath,outputLog)


    #########################
    #Pre-process short reads#
    #########################

    #Map short reads to reference
    outputLog,systemMessage=nPhaseFunctions.shortReadMapping(args.strainName,args.shortReadFile_R1,args.shortReadFile_R2,args.reference,mappedShortReadPath,args.threads)
    print(systemMessage)
    updateLog(fullLogPath,outputLog)

    #Variant call short reads and select SNPs only
    estimatedPloidy="2" #This argument is required for GATK's variant calling and we can expect most SNPs to only have two alleles anyway

    shortReadBam=os.path.join(mappedShortReadPath,args.strainName+".final.bam")
    shortReadVCF=os.path.join(variantCalledShortReadPath,args.strainName+".vcf")
    p=subprocess.run(["gatk","HaplotypeCaller","-R",args.reference,"-ploidy",estimatedPloidy,"-I",shortReadBam,"-O",shortReadVCF],stderr=subprocess.PIPE,stdout=subprocess.PIPE, universal_newlines=True)
    logText="COMMAND: "+" ".join(["gatk","HaplotypeCaller","-R",args.reference,"-ploidy",estimatedPloidy,"-I",shortReadBam,"-O",shortReadVCF])+"\n\n"
    if p.stderr!="" or p.stdout !="":
        logText+="STDERR:\n\n"+p.stderr+"\n\nSTDOUT:\n\n"+p.stdout+"\n\n"
    updateLog(fullLogPath,logText)

    shortReadSNPsVCF=os.path.join(variantCalledShortReadPath,args.strainName+".SNPs.vcf")
    p=subprocess.run(["gatk","SelectVariants","-R",args.reference,"--variant",shortReadVCF,"-O",shortReadSNPsVCF,"--select-type-to-include","SNP"],stderr=subprocess.PIPE,stdout=subprocess.PIPE, universal_newlines=True)
    logText="COMMAND: "+" ".join(["gatk","SelectVariants","-R",args.reference,"--variant",shortReadVCF,"-O",shortReadSNPsVCF,"--select-type-to-include","SNP"])+"\n\n"
    if p.stderr!="" or p.stdout !="":
        logText+="STDERR:\n\n"+p.stderr+"\n\nSTDOUT:\n\n"+p.stdout+"\n\n"
    updateLog(fullLogPath,logText)


    #Extract heterozygous positions from VCF file
    #Ugh this is a whole thing.

    SNPVCFFilePath=os.path.join(variantCalledShortReadPath,args.strainName+".SNPs.vcf")
    hetSNPVCFFilePath=os.path.join(variantCalledShortReadPath,args.strainName+".hetSNPs.vcf")

    hetSNPVCFText=""

    SNPVCFFile=open(SNPVCFFilePath,"r")
    firstFormat=True
    for line in SNPVCFFile:
        if "#" in line:
            hetSNPVCFText+=line
        else:
            if firstFormat:
                format=line.split("\t")[8].split(":")
                if "GT" in format:
                    GTIndex=format.index("GT")
                    firstFormat=False
                else:
                    print("GT not found in FORMAT column of vcf file, cannot identify heterozygous SNPs")
                    exit(0)
            GTData=line.split("\t")[9].split(":")[GTIndex]
            nHaplotypes=len(set(GTData).difference({"|","/"}))
            if nHaplotypes>1:
                hetSNPVCFText+=line

    SNPVCFFile.close()

    hetSNPVCFFile=open(hetSNPVCFFilePath,"w")
    hetSNPVCFFile.write(hetSNPVCFText)
    hetSNPVCFText="" #Just saving some memory since this isn't in a function
    hetSNPVCFFile.close()

    print("Identified heterozygous SNPs in short read VCF")

    shortReadPositionsOutputFilePath=os.path.join(mappedShortReadPath,args.strainName+".hetSNPs.positions.tsv")
    nPhaseFunctions.getShortReadPositions(hetSNPVCFFilePath,shortReadPositionsOutputFilePath)

    shortReadSNPsBedFilePath=os.path.join(mappedShortReadPath,args.strainName+".hetSNPs.bed")

    shortReadSNPsBedText=""

    shortReadPositionsOutputFile=open(shortReadPositionsOutputFilePath,"r")
    for line in shortReadPositionsOutputFile:
        if "#" in line:
            shortReadSNPsBedText+=line
        else:
            line=line.strip("\n")
            line=line.replace(":","\t")
            line=line.split("\t")
            line=[line[0],str(int(line[1])-1),line[1]]
            shortReadSNPsBedText+="\t".join(line)+"\n"

    shortReadPositionsOutputFile.close()

    shortReadSNPsBedFile=open(shortReadSNPsBedFilePath,"w")
    shortReadSNPsBedFile.write(shortReadSNPsBedText)
    shortReadSNPsBedText="" #Just saving some memory since this isn't in a function
    shortReadSNPsBedFile.close()

    print("Extracted heterozygous SNP info based on short read VCF")

    #Reduce long reads to their heterozygous SNPs
    cleanLongReadSamFile=os.path.join(mappedLongReadPath,args.strainName+".sorted.sam")
    minQ=0.01   # Currently
    minMQ=0     # Not
    minAln=0.5  # Used
    longReadPositionNTFile=os.path.join(variantCalledLongReadPath,args.strainName+".hetPositions.SNPxLongReads.tsv")
    nPhaseFunctions.assignLongReadToSNPs(cleanLongReadSamFile,shortReadSNPsBedFilePath,args.reference,minQ,minMQ,minAln,longReadPositionNTFile)

    #Only keep the longest reads and get rid of duplicate heterozygous SNP profiles, compensate by keeping context coverage information.
    minCov=0     # Currently
    minRatio=0   # Not
    minTrioCov=0 # Used
    validatedSNPAssignmentsFile=os.path.join(variantCalledLongReadPath,args.strainName+".hetPositions.SNPxLongReads.validated.tsv")
    contextDepthsFile=os.path.join(overlapPath,args.strainName+".contextDepths.tsv")
    nPhaseFunctions.longReadValidation(longReadPositionNTFile,minCov,minRatio,minTrioCov,validatedSNPAssignmentsFile,contextDepthsFile)

    #Run everything through nPhase
    phaseTool.nPhase(validatedSNPAssignmentsFile,args.strainName,contextDepthsFile,phasedPath,basePath,args.reference,args.minSim,args.minOvl,args.minLen,args.maxID,args.nPairs,int(args.threads))

    readmeText="\nPhased files can be found at "+phasedPath+"\nThe *_variants.tsv file contains information on the consensus heterozygous variants present in each predicted haplotig.\nThe *_clusterReadNames.tsv file contains information on the reads which comprise each cluster."
    print(readmeText)
    updateLog(readmePath,readmeText)

    #Simplify datavis
    dataVisPath=os.path.join(phasedPath,args.strainName+"_"+str(args.minOvl)+"_"+str(args.minSim)+"_"+str(args.maxID)+"_"+str(args.minLen)+"_phasedDataFull.tsv")
    simpleOutPath=os.path.join(phasedPath,args.strainName+"_"+str(args.minOvl)+"_"+str(args.minSim)+"_"+str(args.maxID)+"_"+str(args.minLen)+"_phasedDataSimple.tsv")
    nPhaseFunctions.simplifyDataVis(dataVisPath,simpleOutPath,1000)

    #Generate plots
    datavisPath=os.path.join(datavisFolderPath,args.strainName+"_"+str(args.minOvl)+"_"+str(args.minSim)+"_"+str(args.maxID)+"_"+str(args.minLen)+"_")
    nPhaseFunctions.generatePhasingVis(simpleOutPath,datavisPath)

    #Coverage
    simpleOutPath=os.path.join(phasedPath,args.strainName+"_"+str(args.minOvl)+"_"+str(args.minSim)+"_"+str(args.maxID)+"_"+str(args.minLen)+"_covVis.tsv")
    nPhaseFunctions.generateCoverageVis(simpleOutPath,datavisPath)

    #Generate FastQ Files
    haplotigReadNameFile=os.path.join(phasedPath,args.strainName+"_"+str(args.minOvl)+"_"+str(args.minSim)+"_"+str(args.maxID)+"_"+str(args.minLen)+"_clusterReadNames.tsv")
    fastQFilePrefix=os.path.join(phasedFastqPath,args.strainName+"_"+str(args.minOvl)+"_"+str(args.minSim)+"_"+str(args.maxID)+"_"+str(args.minLen)+"_")

    nPhaseFunctions.generateLongReadFastQFiles(haplotigReadNameFile,args.longReadFile,fastQFilePrefix)#Needs to work more efficiently, shouldn't just load the entire fastq into memory...right? (why not?) Do I do that at any other point? #Make this last just in case.

    readmeText="\nLong reads can be found in "+phasedFastqPath
    print(readmeText)
    updateLog(readmePath,readmeText)

    print("You can consult the readme at "+readmePath+" if you want a bit of guidance about your results. Please raise any issues on https://github.com/nPhasePipeline/nPhase")

    #Discordance
    simpleOutPath=os.path.join(phasedPath,args.strainName+"_"+str(args.minOvl)+"_"+str(args.minSim)+"_"+str(args.maxID)+"_"+str(args.minLen)+"_discordanceVis.tsv")
    try:
        nPhaseFunctions.generateDiscordanceVis(simpleOutPath,datavisPath)
    except:
        print("nPhase was unable to automatically generate the discordance plot (plotnine seems to have issues generating violin plots), the raw data is available in the $prefix_discordanceVis.tsv file in the Phased folder.")

    readmeText="\nPlot can be found at "+datavisFolderPath
    print(readmeText)
    updateLog(readmePath,readmeText)

    return 0


def nPhaseAlgorithm(args):

    basePath=os.path.join(args.outputFolder,args.strainName)
    phasedPath=os.path.join(basePath,"Phased")
    phasedFastqPath=os.path.join(basePath,"Phased","FastQ")
    datavisFolderPath=os.path.join(basePath,"Phased","Plots")
    logPath=os.path.join(basePath,"Logs") #Currently not producing any logs

    allPaths=[basePath,phasedPath,phasedFastqPath,datavisFolderPath,logPath]

    for path in allPaths:
        os.makedirs(path, exist_ok=True)

    readmePath=os.path.join(basePath,"Readme.txt")
    fullLogPath=os.path.join(basePath,"Logs","fullLog.txt")

    #Run everything through nPhase
    phaseTool.nPhase(args.validatedSNPAssignmentsFile,args.strainName,args.contextDepthsFile,phasedPath,basePath,args.reference,args.minSim,args.minOvl,args.minLen,args.maxID,args.nPairs,int(args.threads))

    readmeText="\nPhased files can be found at "+phasedPath+"\nThe *_variants.tsv file contains information on the consensus heterozygous variants present in each predicted haplotig.\nThe *_clusterReadNames.tsv file contains information on the reads which comprise each cluster."
    print(readmeText)
    updateLog(readmePath,readmeText)

    #Simplify datavis
    dataVisPath=os.path.join(phasedPath,args.strainName+"_"+str(args.minOvl)+"_"+str(args.minSim)+"_"+str(args.maxID)+"_"+str(args.minLen)+"_phasedDataFull.tsv")
    simpleOutPath=os.path.join(phasedPath,args.strainName+"_"+str(args.minOvl)+"_"+str(args.minSim)+"_"+str(args.maxID)+"_"+str(args.minLen)+"_phasedDataSimple.tsv")
    nPhaseFunctions.simplifyDataVis(dataVisPath,simpleOutPath,1000)

    #Generate plots
    datavisPath=os.path.join(datavisFolderPath,args.strainName+"_"+str(args.minOvl)+"_"+str(args.minSim)+"_"+str(args.maxID)+"_"+str(args.minLen)+"_")
    nPhaseFunctions.generatePhasingVis(simpleOutPath,datavisPath)

    #Coverage
    simpleOutPath=os.path.join(phasedPath,args.strainName+"_"+str(args.minOvl)+"_"+str(args.minSim)+"_"+str(args.maxID)+"_"+str(args.minLen)+"_covVis.tsv")
    nPhaseFunctions.generateCoverageVis(simpleOutPath,datavisPath)

    #Discordance
    simpleOutPath=os.path.join(phasedPath,args.strainName+"_"+str(args.minOvl)+"_"+str(args.minSim)+"_"+str(args.maxID)+"_"+str(args.minLen)+"_discordanceVis.tsv")
    try:
        nPhaseFunctions.generateDiscordanceVis(simpleOutPath,datavisPath)
    except:
        print("nPhase was unable to automatically generate the discordance plot (likely due to memory issues), the raw data is available in the $prefix_discordanceVis.tsv file in the Phased folder.")

    readmeText="\nPlot can be found at "+datavisFolderPath
    print(readmeText)
    updateLog(readmePath,readmeText)

    #Generate FastQ Files
    haplotigReadNameFile=os.path.join(phasedPath,args.strainName+"_"+str(args.minOvl)+"_"+str(args.minSim)+"_"+str(args.maxID)+"_"+str(args.minLen)+"_clusterReadNames.tsv")
    fastQFilePrefix=os.path.join(phasedFastqPath,args.strainName+"_"+str(args.minOvl)+"_"+str(args.minSim)+"_"+str(args.maxID)+"_"+str(args.minLen)+"_")

    nPhaseFunctions.generateLongReadFastQFiles(haplotigReadNameFile,args.longReadFile,fastQFilePrefix) #Needs to work more efficiently, shouldn't just load the entire fastq into memory...right? (why not?) Do I do that at any other point? #Make this last just in case.

    readmeText="\nLong reads can be found in "+phasedFastqPath
    print(readmeText)
    updateLog(readmePath,readmeText)

    print("You can consult the readme at "+readmePath+" if you want a bit of guidance about your results. Please raise any issues on https://github.com/nPhasePipeline/nPhase")

    return 0

def partialPipeline(args):
    allPaths=[]

    #Create folder structure
    basePath=os.path.join(args.outputFolder,args.strainName)
    mappedShortReadPath=os.path.join(basePath,"Mapped","shortReads")
    if args.mappedLongReads=="noMapped":
        mappedLongReadPath=os.path.join(basePath,"Mapped","longReads")
        allPaths.append(mappedLongReadPath)
    variantCalledShortReadPath=os.path.join(basePath,"VariantCalls","shortReads")
    variantCalledLongReadPath=os.path.join(basePath,"VariantCalls","longReads")
    overlapPath=os.path.join(basePath,"Overlaps")
    phasedPath=os.path.join(basePath,"Phased")
    phasedFastqPath=os.path.join(basePath,"Phased","FastQ")
    datavisFolderPath=os.path.join(basePath,"Phased","Plots")
    logPath=os.path.join(basePath,"Logs") #Currently not producing any logs

    allPaths+=[basePath,mappedShortReadPath,variantCalledShortReadPath,variantCalledLongReadPath,overlapPath,phasedPath,phasedFastqPath,datavisFolderPath,logPath]

    for path in allPaths:
        os.makedirs(path, exist_ok=True)

    readmePath=os.path.join(basePath,"Readme.txt")
    fullLogPath=os.path.join(basePath,"Logs","fullLog.txt")

    ########################
    #Pre-process reference #
    ########################

    #Make sure the reference is indexed (might be missing a process)
    if not os.path.isfile(args.reference+".fai"):
        p=subprocess.run(["samtools","faidx",args.reference],stderr=subprocess.PIPE,stdout=subprocess.PIPE, universal_newlines=True)
        logText="COMMAND: "+" ".join(["samtools","faidx",args.reference])+"\n\n"
        if p.stderr!="" or p.stdout !="":
            logText+="STDERR:\n\n"+p.stderr+"\n\nSTDOUT:\n\n"+p.stdout+"\n\n"
        updateLog(fullLogPath,logText)

    if not os.path.isfile(args.reference+".bwt"):
        p=subprocess.run(["bwa","index",args.reference],stderr=subprocess.PIPE,stdout=subprocess.PIPE, universal_newlines=True)
        logText="COMMAND: "+" ".join(["bwa","index",args.reference])+"\n\n"
        if p.stderr!="" or p.stdout !="":
            logText+="STDERR:\n\n"+p.stderr+"\n\nSTDOUT:\n\n"+p.stdout+"\n\n"
        updateLog(fullLogPath,logText)

    if not os.path.isfile(args.reference+".dict"):
        p=subprocess.run(["gatk","CreateSequenceDictionary","-R",args.reference],stderr=subprocess.PIPE,stdout=subprocess.PIPE, universal_newlines=True)
        logText="COMMAND: "+" ".join(["gatk","CreateSequenceDictionary","-R",args.reference])+"\n\n"
        if p.stderr!="" or p.stdout !="":
            logText+="STDERR:\n\n"+p.stderr+"\n\nSTDOUT:\n\n"+p.stdout+"\n\n"
        updateLog(fullLogPath,logText)

    ########################
    #Pre-process long reads#
    ########################

    #Map long reads to reference
    if args.mappedLongReads=="noMapped":
        splitReadFlag="260" #This flag allows split reads
        outputLog,systemMessage=nPhaseFunctions.longReadMapping(args.strainName,args.longReadFile,args.reference,mappedLongReadPath,splitReadFlag,args.longReadPlatform,args.threads)
        print(systemMessage)
        updateLog(fullLogPath,outputLog)


    #########################
    #Pre-process short reads#
    #########################

    #Map short reads to reference
    if args.mappedShortReads=="noMapped" and args.vcfFile=="noVCF":
        outputLog,systemMessage=nPhaseFunctions.shortReadMapping(args.strainName,args.shortReadFile_R1,args.shortReadFile_R2,args.reference,mappedShortReadPath,args.threads)
        print(systemMessage)
        updateLog(fullLogPath,outputLog)

    if args.vcfFile=="noVCF":
        #Variant call short reads and select SNPs only
        estimatedPloidy="2" #This argument is required for GATK's variant calling and we can expect most SNPs to only have two alleles anyway

        if args.mappedShortReads=="noMapped":
            shortReadBam=os.path.join(mappedShortReadPath,args.strainName+".final.bam")
        else:
            shortReadBam=args.mappedShortReads
        shortReadVCF=os.path.join(variantCalledShortReadPath,args.strainName+".vcf")
        p=subprocess.run(["gatk","HaplotypeCaller","-R",args.reference,"-ploidy",estimatedPloidy,"-I",shortReadBam,"-O",shortReadVCF],stderr=subprocess.PIPE,stdout=subprocess.PIPE, universal_newlines=True)
        logText="COMMAND: "+" ".join(["gatk","HaplotypeCaller","-R",args.reference,"-ploidy",estimatedPloidy,"-I",shortReadBam,"-O",shortReadVCF])+"\n\n"
        if p.stderr!="" or p.stdout !="":
            logText+="STDERR:\n\n"+p.stderr+"\n\nSTDOUT:\n\n"+p.stdout+"\n\n"
        updateLog(fullLogPath,logText)
    else:
        shortReadVCF=args.vcfFile

    shortReadSNPsVCF=os.path.join(variantCalledShortReadPath,args.strainName+".SNPs.vcf")
    p=subprocess.run(["gatk","SelectVariants","-R",args.reference,"--variant",shortReadVCF,"-O",shortReadSNPsVCF,"--select-type-to-include","SNP"],stderr=subprocess.PIPE,stdout=subprocess.PIPE, universal_newlines=True)
    logText="COMMAND: "+" ".join(["gatk","SelectVariants","-R",args.reference,"--variant",shortReadVCF,"-O",shortReadSNPsVCF,"--select-type-to-include","SNP"])+"\n\n"
    if p.stderr!="" or p.stdout !="":
        logText+="STDERR:\n\n"+p.stderr+"\n\nSTDOUT:\n\n"+p.stdout+"\n\n"
    updateLog(fullLogPath,logText)


    #Extract heterozygous positions from VCF file
    #Ugh this is a whole thing.

    SNPVCFFilePath=os.path.join(variantCalledShortReadPath,args.strainName+".SNPs.vcf")
    hetSNPVCFFilePath=os.path.join(variantCalledShortReadPath,args.strainName+".hetSNPs.vcf")

    hetSNPVCFText=""

    SNPVCFFile=open(SNPVCFFilePath,"r")
    firstFormat=True
    for line in SNPVCFFile:
        if "#" in line:
            hetSNPVCFText+=line
        else:
            if firstFormat:
                format=line.split("\t")[8].split(":")
                if "GT" in format:
                    GTIndex=format.index("GT")
                    firstFormat=False
                else:
                    print("GT not found in FORMAT column of vcf file, cannot identify heterozygous SNPs")
                    exit(0)
            GTData=line.split("\t")[9].split(":")[GTIndex]
            nHaplotypes=len(set(GTData).difference({"|","/"}))
            if nHaplotypes>1:
                hetSNPVCFText+=line

    SNPVCFFile.close()

    hetSNPVCFFile=open(hetSNPVCFFilePath,"w")
    hetSNPVCFFile.write(hetSNPVCFText)
    hetSNPVCFText="" #Just saving some memory since this isn't in a function
    hetSNPVCFFile.close()

    print("Identified heterozygous SNPs in short read VCF")

    shortReadPositionsOutputFilePath=os.path.join(mappedShortReadPath,args.strainName+".hetSNPs.positions.tsv")
    nPhaseFunctions.getShortReadPositions(hetSNPVCFFilePath,shortReadPositionsOutputFilePath)

    shortReadSNPsBedFilePath=os.path.join(mappedShortReadPath,args.strainName+".hetSNPs.bed")

    shortReadSNPsBedText=""

    shortReadPositionsOutputFile=open(shortReadPositionsOutputFilePath,"r")
    for line in shortReadPositionsOutputFile:
        if "#" in line:
            shortReadSNPsBedText+=line
        else:
            line=line.strip("\n")
            line=line.replace(":","\t")
            line=line.split("\t")
            line=[line[0],str(int(line[1])-1),line[1]]
            shortReadSNPsBedText+="\t".join(line)+"\n"

    shortReadPositionsOutputFile.close()

    shortReadSNPsBedFile=open(shortReadSNPsBedFilePath,"w")
    shortReadSNPsBedFile.write(shortReadSNPsBedText)
    shortReadSNPsBedText="" #Just saving some memory since this isn't in a function
    shortReadSNPsBedFile.close()

    print("Extracted heterozygous SNP info based on short read VCF")

    #Reduce long reads to their heterozygous SNPs
    if args.mappedLongReads=="noMapped":
        cleanLongReadSamFile=os.path.join(mappedLongReadPath,args.strainName+".sorted.sam")
    else:
        cleanLongReadSamFile=args.mappedLongReads
    minQ=0.01   # Currently
    minMQ=0     # Not
    minAln=0.5  # Used
    longReadPositionNTFile=os.path.join(variantCalledLongReadPath,args.strainName+".hetPositions.SNPxLongReads.tsv")
    nPhaseFunctions.assignLongReadToSNPs(cleanLongReadSamFile,shortReadSNPsBedFilePath,args.reference,minQ,minMQ,minAln,longReadPositionNTFile)

    #Only keep the longest reads and get rid of duplicate heterozygous SNP profiles, compensate by keeping context coverage information.
    minCov=0     # Currently
    minRatio=0   # Not
    minTrioCov=0 # Used
    validatedSNPAssignmentsFile=os.path.join(variantCalledLongReadPath,args.strainName+".hetPositions.SNPxLongReads.validated.tsv")
    contextDepthsFile=os.path.join(overlapPath,args.strainName+".contextDepths.tsv")
    nPhaseFunctions.longReadValidation(longReadPositionNTFile,minCov,minRatio,minTrioCov,validatedSNPAssignmentsFile,contextDepthsFile)

    #Run everything through nPhase
    phaseTool.nPhase(validatedSNPAssignmentsFile,args.strainName,contextDepthsFile,phasedPath,basePath,args.reference,args.minSim,args.minOvl,args.minLen,args.maxID,args.nPairs,int(args.threads))

    readmeText="\nPhased files can be found at "+phasedPath+"\nThe *_variants.tsv file contains information on the consensus heterozygous variants present in each predicted haplotig.\nThe *_clusterReadNames.tsv file contains information on the reads which comprise each cluster."
    print(readmeText)
    updateLog(readmePath,readmeText)

    #Simplify datavis
    dataVisPath=os.path.join(phasedPath,args.strainName+"_"+str(args.minOvl)+"_"+str(args.minSim)+"_"+str(args.maxID)+"_"+str(args.minLen)+"_phasedDataFull.tsv")
    simpleOutPath=os.path.join(phasedPath,args.strainName+"_"+str(args.minOvl)+"_"+str(args.minSim)+"_"+str(args.maxID)+"_"+str(args.minLen)+"_phasedDataSimple.tsv")
    nPhaseFunctions.simplifyDataVis(dataVisPath,simpleOutPath,1000)

    ################
    #Generate plots#
    ################

    #Phased
    datavisPath=os.path.join(datavisFolderPath,args.strainName+"_"+str(args.minOvl)+"_"+str(args.minSim)+"_"+str(args.maxID)+"_"+str(args.minLen)+"_")
    nPhaseFunctions.generatePhasingVis(simpleOutPath,datavisPath)

    #Coverage
    simpleOutPath=os.path.join(phasedPath,args.strainName+"_"+str(args.minOvl)+"_"+str(args.minSim)+"_"+str(args.maxID)+"_"+str(args.minLen)+"_covVis.tsv")
    nPhaseFunctions.generateCoverageVis(simpleOutPath,datavisPath)

    #Discordance
    simpleOutPath=os.path.join(phasedPath,args.strainName+"_"+str(args.minOvl)+"_"+str(args.minSim)+"_"+str(args.maxID)+"_"+str(args.minLen)+"_discordanceVis.tsv")
    try:
        nPhaseFunctions.generateDiscordanceVis(simpleOutPath,datavisPath)
    except:
        print("nPhase was unable to automatically generate the discordance plot (likely due to memory issues), the raw data is available in the $prefix_discordanceVis.tsv file in the Phased folder.")

    readmeText="\nPlot can be found at "+datavisFolderPath
    print(readmeText)
    updateLog(readmePath,readmeText)


    #Generate FastQ Files
    haplotigReadNameFile=os.path.join(phasedPath,args.strainName+"_"+str(args.minOvl)+"_"+str(args.minSim)+"_"+str(args.maxID)+"_"+str(args.minLen)+"_clusterReadNames.tsv")
    fastQFilePrefix=os.path.join(phasedFastqPath,args.strainName+"_"+str(args.minOvl)+"_"+str(args.minSim)+"_"+str(args.maxID)+"_"+str(args.minLen)+"_")

    nPhaseFunctions.generateLongReadFastQFiles(haplotigReadNameFile,args.longReadFile,fastQFilePrefix)#Needs to work more efficiently, shouldn't just load the entire fastq into memory...right? (why not?) Do I do that at any other point? #Make this last just in case.

    readmeText="\nLong reads can be found in "+phasedFastqPath
    print(readmeText)
    updateLog(readmePath,readmeText)

    print("You can consult the readme at "+readmePath+" if you want a bit of guidance about your results. Please raise any issues on https://github.com/nPhasePipeline/nPhase")

    return 0

def cleaning(args):
    cleaningTool.nPhaseCleaning(args.phasingResult,args.longReadFile,args.strainName,args.percentKept,args.maxDiscordance,args.deduplicate,args.FFBool)

def main():
    parser=argparse.ArgumentParser(prog="nPhase",description='Full ploidy agnostic phasing pipeline',add_help=False)
    parser.add_argument('--version',action='version',version='%(prog)s 1.2.0')

    #CREATING SEPARATE MODES
    subparsers = parser.add_subparsers(help="selecting 'pipeline' will run through all of the steps, 'nPhase' will only perform the phasing operation, 'partial' will only perform part of the pipeline, 'cleaning' will run automated cleaning steps on nPhase results",dest='command')

    parser_a = subparsers.add_parser('pipeline', help='Run the entire nPhase pipeline on your sample.',parents=[parser])
    required_a=parser_a.add_argument_group('required arguments')

    parser_b = subparsers.add_parser('algorithm', help='Only run the nPhase algorithm. NOTE: This will require files generated by running the pipeline mode',parents=[parser])
    required_b=parser_b.add_argument_group('required arguments')

    parser_c = subparsers.add_parser('partial', help='Only run parts of the nPhase pipeline on your sample. NOTE: if you run into any issues with this mode please use the pipeline mode instead.',parents=[parser])
    required_c=parser_c.add_argument_group('required arguments')

    cleaningParser=subparsers.add_parser('cleaning',help='Clean nPhase results using a simple three step pipeline. Filtering, merging, gap filling.',parents=[parser])
    required_Cleaning=cleaningParser.add_argument_group('required arguments')

    #SHARED REQUIRED ARGUMENTS

    for parserReference in [required_a,required_b,required_c,required_Cleaning]:
        parserReference.add_argument('--sampleName',required=True, dest='strainName',help='Name of your sample, ex: "Individual_1"')
        parserReference.add_argument('--longReads',required=True, dest='longReadFile',help='Path to long read FastQ file, ex: /home/longReads/Individual_1.fastq.gz')

    #FULL PIPELINE SPECIFIC ARGUMENTS
    required_a.add_argument('--longReadPlatform',required=True, dest='longReadPlatform',choices=['ont','pacbio'],help="Long read platform, must be 'ont' or 'pacbio'")
    required_a.add_argument('--R1',required=True, dest='shortReadFile_R1',help='Path to paired end short read FastQ file #1, ex: /home/shortReads/Individual_1_R1.fastq.gz')
    required_a.add_argument('--R2',required=True, dest='shortReadFile_R2',help='Path to paired end short read FastQ file #2, ex: /home/shortReads/Individual_1_R2.fastq.gz')
    required_a.add_argument('--nPairs',dest='nPairs',type=int,nargs="?",default=1,help='Number of overlapping clusters to merge per iterative step in the nPhase algorithm. Higher values can speed up runtime at the cost of accuracy. Default 1')

    #NPHASE SPECIFIC ARGUMENTS
    required_b.add_argument('--contextDepth',required=True, dest='contextDepthsFile',help='Path to context depths file, ex: /home/phased/Individual_1/Overlaps/Individual_1.contextDepths.tsv')
    required_b.add_argument('--processedLongReads',required=True, dest='validatedSNPAssignmentsFile',help='Path to validated long read SNPs, ex: /home/phased/Individual_1/VariantCalls/longReads/Individual_1.hetPositions.SNPxLongReads.validated.tsv')
    required_b.add_argument('--nPairs',dest='nPairs',type=int,nargs="?",default=1,help='Number of overlapping clusters to merge per iterative step in the nPhase algorithm. Higher values can speed up runtime at the cost of accuracy. Default 1')

    #PARTIAL PIPELINE SPECIFIC ARGUMENTS
    required_c.add_argument('--mappedShortReads', dest='mappedShortReads', default="noMapped",help='Path to mapped, sorted short read file in BAM format, ex: /home/phased/Individual_1/Mapped/shortReads/Individual_1.final.bam')
    required_c.add_argument('--vcf',dest='vcfFile',default="noVCF",help="Path to VCF file (must be generated with GATK --ploidy=2), ex:/home/phased/Individual_1/VariantCalls/shortReads/Individual.vcf")
    required_c.add_argument('--mappedLongReads', dest='mappedLongReads', default="noMapped",help='Path to mapped long read file in SAM format and should be filtered to keep split reads (flag 260), ex: /home/phased/Individual_1/Mapped/longReads/Individual_1.sorted.sam')
    required_c.add_argument('--longReadPlatform', dest='longReadPlatform',default="noLRMapping",choices=['ont','pacbio'],help="Long read platform, must be 'ont' or 'pacbio'")
    required_c.add_argument('--R1', dest='shortReadFile_R1',default="noSRMapping",help='Path to paired end short read FastQ file #1, ex: /home/shortReads/Individual_1_R1.fastq.gz')
    required_c.add_argument('--R2', dest='shortReadFile_R2',default="noSRMapping",help='Path to paired end short read FastQ file #2, ex: /home/shortReads/Individual_1_R2.fastq.gz')
    required_c.add_argument('--nPairs',dest='nPairs',type=int,nargs="?",default=1,help='Number of overlapping clusters to merge per iterative step in the nPhase algorithm. Higher values can speed up runtime at the cost of accuracy. Default 1')

    #CLEANING SPECIFIC ARGUMENTS
    required_Cleaning.add_argument('--resultFolder',required=True,dest='phasingResult',help='Path to output folder, ex: /home/nphaseResults/Individual_1/')
    required_Cleaning.add_argument('--filterPct', dest='percentKept',type=float,nargs="?",default=2.0,help='Percentage of results to filter out. Default 2.0')
    required_Cleaning.add_argument('--fillGaps',dest='deduplicate',type=bool,nargs="?",default=True,help='Attempt to fill gaps by redistributing reads from most covered cluster. Set to 0 to disable this step. Default 1')
    required_Cleaning.add_argument('--filterFirst',dest='FFBool',type=bool,nargs="?",default=False,help='Perform the filtering step before the merging step. Set to 1 to enable. Default 0')
    required_Cleaning.add_argument('--setMaxDiscordance',dest='maxDiscordance',type=float,nargs="?",default=0.0,help='If this is set higher than 0, use this percentage as the stopping rule for merging instead of using the mean discordance of raw results. Inputting 6.8 means 6.8%% discordance max allowed. Default 0')

    #OPTIONAL SHARED ARGUMENTS
    for parserReference in [required_a,required_b,required_c]:
        parserReference.add_argument('--reference',required=True,dest='reference',help='Path to fasta file of reference genome to align to, ex: /home/reference/Individual_reference.fasta')
        parserReference.add_argument('--output',required=True,dest='outputFolder',help='Path to output folder, ex: /home/phased/')
        parserReference.add_argument('--threads',dest='threads',type=str,nargs="?",default="8",help='Number of threads to use on some steps. Default 8')
        parserReference.add_argument('--maxID',dest='maxID',type=float,nargs="?",default=0.05,help='MaxID parameter, determines how different two clusters must be to prevent them from merging. Default 0.05')
        parserReference.add_argument('--minOvl',dest='minOvl',type=float,nargs="?",default=0.1,help='minOvl parameter, determines the minimal percentage of overlap required to allow a merge between two clusters that have fewer than 100 heterozygous SNPs in common. Default 0.1')
        parserReference.add_argument('--minSim',dest='minSim',type=float,nargs="?",default=0.01,help='minSim parameter, determines the minimal percentage of similarity required to allow a merge between two clusters. Default 0.01')
        parserReference.add_argument('--minLen',dest='minLen',type=int,nargs="?",default=0,help='minLen parameter, any cluster based on fewer than N reads will not be output. Default 0')

    args, unknown = parser.parse_known_args()

    if "mappedShortReads" in dir(args) or "vcf" in dir(args) or "mappedLongReads" in dir(args):
        partialPipeline(args)
    elif "longReadPlatform" in dir(args):
        nPhasePipeline(args)
    elif "contextDepthsFile" in dir(args):
        nPhaseAlgorithm(args)
    elif "phasingResult" in dir(args):
        cleaning(args)
    else:
        print("Select which mode of nphase to use. For example:\n\nnphase pipeline [args] [-h]\n\nnphase algorithm [args] [-h]\n\nnphase partial [args] [-h]\n\nnphase filter [args] [-h]")

    exit(0)

if __name__=="__main__":
    main()
    exit(0)





import itertools
import sortedcontainers
import random
import os
import glob
import bin.nPhasePipelineFunctions as nPhaseFunctions

def nPhaseCleaning(nPhaseResultFolder,longReadFilePath,strainPrefix,percentKept=98,maxDiscordance=0,deduplicate=True,FFBool=False):
    percentKept=1-(percentKept/100)
    
    readClusterPath=glob.glob(os.path.join(nPhaseResultFolder,"Phased","*clusterReadNames.tsv"))[0]
    knownHetFile=glob.glob(os.path.join(nPhaseResultFolder,"VariantCalls","shortReads","*.hetSNPs.vcf"))[0]
    readSNPsPath=glob.glob(os.path.join(nPhaseResultFolder,"VariantCalls","longReads","*.hetPositions.SNPxLongReads.validated.tsv"))[0]
    outPath=os.path.join(nPhaseResultFolder,"Phased","Cleaned",strainPrefix)
    os.makedirs(outPath,exist_ok=True)
    os.makedirs(os.path.join(outPath,"cleanFastQ"),exist_ok=True)
    os.makedirs(os.path.join(outPath,"Plots"),exist_ok=True)
    
    phasedSNPFilePath=os.path.join(outPath,strainPrefix+"_cleaned.variants.tsv")
    phasedClusterPath=os.path.join(outPath,strainPrefix+"_cleaned.clusterReadNames.tsv")

    clusterReads=readClusterReads(readClusterPath)
    readSNPs=readReadSNPs(readSNPsPath)
    
    clusters=rebuildClusters(clusterReads,readSNPs)
    print("Number of clusters before cleaning:",len(clusters))
    
    if FFBool:
        if percentKept<1.0:
            stitchedClusters=filterLowCoverageClusters(clusters,percentKept)
            stitchedClusters=iterativeStitching(stitchedClusters,maxDiscordance)
        else:
            stitchedClusters=iterativeStitching(clusters,maxDiscordance)
    else:
        stitchedClusters=iterativeStitching(clusters,maxDiscordance)
        if percentKept<1.0:
            stitchedClusters=filterLowCoverageClusters(stitchedClusters,percentKept)

    if deduplicate:
        stitchedClusters=identifyGaps(stitchedClusters,readSNPs,50)
    
    makeVariantsFile(stitchedClusters,phasedSNPFilePath)
    makeClusterReadFile(stitchedClusters,phasedClusterPath)
    
    simplerStitchedClusters={}
    for clusterName,clusterData in stitchedClusters.items():
        simplerStitchedClusters[clusterName]=clusterData["SNPs"]
    
    dataText=nPhaseFunctions.giveMeFullData(simplerStitchedClusters)
    dataVisPath=os.path.join(outPath,strainPrefix+"_phasedDataFull.tsv")
    dataVisFile=open(dataVisPath,"w")
    dataVisFile.write(dataText)
    dataVisFile.close()
    
    simpleOutPath=os.path.join(outPath,strainPrefix+"_phasedDataSimple.tsv")
    nPhaseFunctions.simplifyDataVis(dataVisPath,simpleOutPath,1000)
    datavisPath=os.path.join(outPath,"Plots",strainPrefix+"_")
    nPhaseFunctions.generatePhasingVis(simpleOutPath,datavisPath)
    
    simpleOutPath=os.path.join(outPath,strainPrefix+"_covVis.tsv")
    windowSize=10000
    minCov=nPhaseFunctions.generateCoverage(phasedClusterPath,readSNPsPath,simpleOutPath,windowSize)
    
    nPhaseFunctions.generateCoverageVis(simpleOutPath,datavisPath)
    
    #Discordance
    simpleOutPath=os.path.join(outPath,strainPrefix+"_discordanceVis.tsv")
    nPhaseFunctions.generateDiscordance(phasedClusterPath,readSNPsPath,simpleOutPath)
    
    nPhaseFunctions.generateDiscordanceVis(simpleOutPath,datavisPath)
    
    fastQOut=os.path.join(outPath,"cleanFastQ",strainPrefix)
    nPhaseFunctions.generateLongReadFastQFiles(phasedClusterPath,longReadFilePath,fastQOut)
    
    print("Cleaning done! Your cleaning results are in",outPath)


def readClusterReads(filePath):
    clusterReadFile=open(filePath,"r")

    readClusterDict={}

    for line in clusterReadFile:
        line=line.strip("\n")
        line=line.split("\t")
        readClusterDict.setdefault(line[0],set())
        readClusterDict[line[0]].add(line[1])

    return readClusterDict

def readReadSNPs(filePath):
    readSNPFile=open(filePath,"r")

    readSNPDict={}

    for line in readSNPFile:
        line=line.strip("\n")
        line=line.split("\t")
        readID=line[0]
        for SNP in line[1:]:
            readSNPDict.setdefault(readID,set())
            readSNPDict[readID].add(SNP)

    return readSNPDict

def rebuildClusters(readClusterDict,readSNPDict):
    rebuiltClusters={}

    for cluster,reads in readClusterDict.items():
        rebuiltClusters.setdefault(cluster,{"SNPs":{},"Reads":set()})
        for read in reads:
            for SNP in readSNPDict[read]:
                rebuiltClusters[cluster]["SNPs"].setdefault(SNP,0)
                rebuiltClusters[cluster]["SNPs"][SNP]+=1
                rebuiltClusters[cluster]["Reads"].add(read)

    return rebuiltClusters

def mergeClusters(clusterA,clusterB):
    mergedCluster={"SNPs":{},"Reads":set()}

    for cluster in [clusterA,clusterB]:
        for SNP,coverage in cluster["SNPs"].items():
            mergedCluster["SNPs"].setdefault(SNP,0)
            mergedCluster["SNPs"][SNP]+=coverage
        mergedCluster["Reads"].update(cluster["Reads"])

    return mergedCluster


def calculateDiscordance(cluster):
    discordanceScore=0  #As a %age in the end, but weighted by coverage for each pos
    totalScore=0

    discordanceDict={}

    for SNP,coverage in cluster.items():
        pos=SNP.split("=")[0]
        base=SNP.split("=")[1]
        discordanceDict.setdefault(pos,{})
        discordanceDict[pos].setdefault(base,0)
        discordanceDict[pos][base]+=coverage

    for position,bases in discordanceDict.items():
        total=0
        highest=0
        for base,coverage in bases.items():
            total+=coverage
            if coverage>highest:
                highest=coverage
        baseDiscordance=total-highest
        discordanceScore+=baseDiscordance
        totalScore+=total

    return discordanceScore/totalScore

def getMeanDiscordance(clusters):
    totalClusters=len(clusters)
    totalDiscordance=0
    for clusterName in clusters:
        totalDiscordance+=calculateDiscordance(clusters[clusterName]["SNPs"])
    return totalDiscordance/totalClusters

def identifyOverlap(clusterA,clusterB):
    clusterAPositions=set([SNP.split("=")[0] for SNP in clusterA.keys()])
    clusterBPositions=set([SNP.split("=")[0] for SNP in clusterB.keys()])
    return clusterAPositions&clusterBPositions

def iterativeStitching(clusters,maxDiscordance):
    readClusters={}
    minOverlap=10
    clusterNumber=1

    meanDiscordance=getMeanDiscordance(clusters)
    print("Mean discordance is",meanDiscordance)

    if maxDiscordance==0:
        maxDiscordance=meanDiscordance

    print("Initializing similarity index")

    similarityIndex=sortedcontainers.SortedList()

    for clusterA,clusterB in itertools.combinations(clusters,2):
        overlappingRegion=identifyOverlap(clusters[clusterA]["SNPs"],clusters[clusterB]["SNPs"])
        overlapLength=len(overlappingRegion)
        if overlapLength>minOverlap:
            mergedCluster=mergeClusters(clusters[clusterA],clusters[clusterB])
            score=calculateDiscordance(mergedCluster["SNPs"])
            if score<maxDiscordance:
                similarityIndex.add([score,-overlapLength,(clusterA,clusterB)])

    print("Similarity index ready")

    print("#Number of clusters\tMerged discordance level\tInitial mean discordance\tDistance from mean\tMerged overlap length")

    while len(similarityIndex)>0:
        highestOverlapLength=similarityIndex[0][1]
        lowestScore=similarityIndex[0][0]
        lowestPair=similarityIndex[0][2]
        if lowestScore<maxDiscordance:
            clusters["stitched_"+str(clusterNumber)]=mergeClusters(clusters[lowestPair[0]],clusters[lowestPair[1]])
            del clusters[lowestPair[0]]
            del clusters[lowestPair[1]]
            getRidOf=[]
            for pair in similarityIndex:
                if lowestPair[0] in pair[2] or lowestPair[1] in pair[2]:
                    getRidOf.append(pair)
            for pair in getRidOf:
                similarityIndex.discard(pair)
            for cluster in clusters.keys():
                if cluster!="stitched_"+str(clusterNumber):
                    overlappingRegion=identifyOverlap(clusters[cluster]["SNPs"],
                                                      clusters["stitched_"+str(clusterNumber)]["SNPs"])
                    overlapLength=len(overlappingRegion)
                    if overlapLength>minOverlap:
                        newMergedCluster=mergeClusters(clusters[cluster],clusters["stitched_"+str(clusterNumber)])
                        newScore=calculateDiscordance(newMergedCluster["SNPs"])
                        if newScore<maxDiscordance:
                            similarityIndex.add([newScore,-overlapLength,(cluster,"stitched_"+str(clusterNumber))])  ##
            clusterNumber+=1
        print(len(clusters),round(lowestScore*100,2),round(meanDiscordance*100,2),
              round(meanDiscordance*100-lowestScore*100,2),-highestOverlapLength)

    print("Merging step over")

    return clusters

def filterLowCoverageClusters(clusters,percentKept):
    #calculate coverage per cluster
    clusterCoverageList=[]
    totalCoverage=0
    for cluster in clusters.keys():
        clusterCoverage=0
        for SNP,coverage in clusters[cluster]["SNPs"].items():  #Maybe only count reads that support the consensus?
            clusterCoverage+=coverage
        #calculate total coverage
        totalCoverage+=clusterCoverage
        clusterCoverageList.append((clusterCoverage,cluster))
    #sort clusters by coverage
    clusterCoverageList.sort(reverse=True)
    #take most covered clusters until you reach 95% of total coverage
    keepN=percentKept*totalCoverage
    filteredClusters={}
    filteredCoverage=0
    for pair in clusterCoverageList:
        cov=pair[0]
        clust=pair[1]
        if filteredCoverage<keepN:
            filteredClusters[clust]=clusters[clust]
            filteredCoverage+=cov
        else:
            return filteredClusters

def identifyGaps(filteredClusters,readSNPs,minGapLen):
    redistributedClusters={}
    #Start by identifying all positions and sort them by chromosome
    genomeCoverage={}
    for cluster in filteredClusters.keys():
        theseSNPs=filteredClusters[cluster]["SNPs"].keys()
        uniquePositions=set([SNP.split("=")[0] for SNP in theseSNPs])
        for pos in uniquePositions:
            chr=pos.split(":")[0]
            pos=int(pos.split(":")[1])
            genomeCoverage.setdefault(chr,{})
            genomeCoverage[chr].setdefault(pos,0)
            genomeCoverage[chr][pos]+=1
    #What is a gap?
    for chromosome in genomeCoverage.keys():
        redistributedClusters.setdefault(chromosome,{})
        chromosomePositions=[]
        for pos,cov in genomeCoverage[chromosome].items():
            chromosomePositions.append((pos,cov))
        chromosomePositions.sort()
        #Identify a gap in here
        ploidy=round(sum([cov for pos,cov in chromosomePositions])/len(chromosomePositions))
        gapPositionDict={}
        i=1
        for position,cov in chromosomePositions:
            if cov<ploidy:
                gapPositionDict.setdefault("gap_"+str(i),set())
                gapPositionDict["gap_"+str(i)].add(chromosome+":"+str(position))
            else:
                i+=1
        tinyGapNames=set()
        for gap in gapPositionDict.keys():
            gapLength=len(gapPositionDict[gap])
            if gapLength<minGapLen:
                tinyGapNames.add(gap)
        for tinyGap in tinyGapNames:
            del gapPositionDict[tinyGap]

        for gap,allGaps in gapPositionDict.items():
            allGaps=[int(SNP.split(":")[1]) for SNP in allGaps]
            allGaps.sort()

        #Okay now we have gaps of adequate size
        #Extract the clusters in this region and their coverage levels.
        clusterPositionDict={}
        for cluster in filteredClusters.keys():
            clusterPositions=[]
            for SNP,coverage in filteredClusters[cluster]["SNPs"].items():
                clusterPositions.append(SNP.split("=")[0])
            clusterPositionDict[cluster]=set(clusterPositions)
        #Now identify, for each gap, every single cluster that overlaps with it.
        gapOverlapDict={}
        for gap,gapPositions in gapPositionDict.items():
            gapOverlapDict.setdefault(gap,[])
            mergeCandidates=[]
            for cluster,clusterPositions in clusterPositionDict.items():
                overlap=gapPositions&clusterPositions
                if overlap!=set():
                    gapOverlapDict[gap].append((len(overlap),cluster))
                if overlap==set() and list(clusterPositions)[0].split(":")[
                    0]==chromosome:  #If the cluster's in the right chromosome, doesn't overlap the gap
                    mergeCandidates.append((len(overlap),cluster))
            gapOverlapDict[gap].sort(reverse=True)
            mergeCandidates.sort(reverse=True)
            #Now, for each cluster that overlaps a gap, let's try redistributing reads.
            #1, identify which of the donor clusters has the highest coverage value for the region
            highestCoverage=0
            highestCandidate=None
            for overlapLen,donorCandidate in gapOverlapDict[gap]:
                if overlapLen>0:
                    coverage=getRegionCoverage(filteredClusters[donorCandidate]["SNPs"],gapPositions)
                    if coverage>highestCoverage:
                        highestCoverage=coverage
                        highestCandidate=donorCandidate
            #1.5, identify the reads of that higher coverage cluster
            includedReads,remainingReads=getRegionReads(filteredClusters[highestCandidate],gapPositions,readSNPs)
            #2, split the reads in that cluster and region into at least two different clusters (and minimize discordance with 2 groups)
            if len(includedReads)>20 and len(gapOverlapDict[gap])>0 and len(mergeCandidates)>0:
                del filteredClusters[highestCandidate]
                clusterA,clusterB,readsA,readsB=minimizeRegionDiscordance(includedReads,readSNPs)

                #3, figure out the lowest discordance merge you can do (with the other clusters of the chromosome
                #Make sure you don't double count reads when you overlap several gaps.

                #Identify which of clusterA and clusterB merges best back into the original cluster
                levelA,mergedA=checkDiscordance(readsA|remainingReads,readSNPs)
                levelB,mergedB=checkDiscordance(readsB|remainingReads,readSNPs)

                if levelA<levelB:
                    filteredClusters[highestCandidate]=mergedA
                    donorCluster=makeCluster(readsB,readSNPs)  #Or just clusterB
                else:
                    filteredClusters[highestCandidate]=mergedB
                    donorCluster=makeCluster(readsA,readSNPs)  #Or just clusterA

                #Replace the donor region with the reads that were chosen

                lowestDiscordance=1000
                lowestCandidate=None
                lowestCluster=None

                for candidateLength,candidate in mergeCandidates:
                    #Test every combination of candidate and remaining cluster

                    #Calculate discordance
                    readClusterToTest=donorCluster["Reads"]|filteredClusters[candidate]["Reads"]
                    testDiscordance,mergedDonor=checkDiscordance(readClusterToTest,readSNPs)
                    if testDiscordance<lowestDiscordance:
                        lowestDiscordance=testDiscordance
                        lowestCandidate=candidate
                        lowestCluster=mergedDonor

                #Keep the one with the lowest discordance
                #Merge the one with the lowest discordance
                del filteredClusters[lowestCandidate]
                filteredClusters[lowestCandidate]=lowestCluster

            else:
                pass
    return filteredClusters

def checkDiscordance(readsFound,readSNPs):
    SNPCluster={"Reads":set(),"SNPs":{}}
    for read in readsFound:
        for SNP in readSNPs[read]:
            SNPCluster["SNPs"].setdefault(SNP,0)
            SNPCluster["SNPs"][SNP]+=1
        SNPCluster["Reads"].add(read)
    return calculateDiscordance(SNPCluster["SNPs"]),SNPCluster

def getRegionReads(cluster,allowedPositions,readSNPs):
    includedReads=set()
    remainingReads=set()
    for read in cluster["Reads"]:
        allSNPs=list(readSNPs[read])
        allPositions=set([SNP.split("=")[0] for SNP in allSNPs])
        if len(allPositions&allowedPositions)>5:
            includedReads.add(read)
        else:
            remainingReads.add(read)
    return includedReads,remainingReads

def makeCluster(reads,readSNPs):
    SNPCluster={"Reads":set(),"SNPs":{}}
    for read in reads:
        for SNP in readSNPs[read]:
            SNPCluster["SNPs"].setdefault(SNP,0)
            SNPCluster["SNPs"][SNP]+=1
        SNPCluster["Reads"].add(read)
    return SNPCluster

def minimizeRegionDiscordance(includedReads,readSNPs):
    startingCluster={}
    consensusSNPs=set()

    #Identify one read that feels like it belongs to cluster #2
    for read in includedReads:
        for SNP in readSNPs[read]:
            pos=SNP.split("=")[0]
            base=SNP.split("=")[1]
            startingCluster.setdefault(pos,{})
            startingCluster[pos].setdefault(base,0)
            startingCluster[pos][base]+=1

    for position,positionData in startingCluster.items():
        maxCov=0
        maxBase=None
        for possibleBase,cov in positionData.items():
            if cov>maxCov:
                maxCov=cov
                maxBase=possibleBase
        consensusSNPs.add(position+"="+maxBase)

    #Identify one read that feels like it belongs to cluster #2
    #Find two reads that disagree the most (even if they disagree very little)
    maxNotInCommon=0
    maxPair=(None,None)

    for readA,readB in itertools.combinations(includedReads,2):
        SNPsA=readSNPs[readA]
        SNPsB=readSNPs[readB]
        posA=set([SNP.split("=")[0] for SNP in SNPsA])
        posB=set([SNP.split("=")[0] for SNP in SNPsB])
        commonPos=posA&posB
        commonSNPs=SNPsA&SNPsB
        SNPsNotInCommon=len(commonPos)-len(commonSNPs)
        if SNPsNotInCommon>=maxNotInCommon:
            maxNotInCommon=SNPsNotInCommon
            maxPair=(readA,readB)
    #Grow both clusters
    clusterA={}
    clusterB={}
    alreadyUsed={maxPair[0],maxPair[1]}
    for SNP in readSNPs[maxPair[0]]:
        pos=SNP.split("=")[0]
        base=SNP.split("=")[1]
        clusterA.setdefault(pos,{})
        clusterA[pos].setdefault(base,0)
        clusterA[pos][base]+=1

    readsA=set()
    readsB=set()

    while len(alreadyUsed)!=len(includedReads):
        clusterAConsensus=getConsensus(clusterA)
        clusterBConsensus=getConsensus(clusterB)
        bestA=None
        bestAScore=0
        bestB=None
        bestBScore=0
        for read in includedReads:
            if read not in alreadyUsed:
                candidateSNPs=readSNPs[read]
                aScore=len(candidateSNPs&clusterAConsensus)
                if aScore>=bestAScore:
                    bestAScore=aScore
                    bestA=read
                bScore=len(candidateSNPs&clusterBConsensus)
                if bScore>=bestBScore:
                    bestBScore=bScore
                    bestB=read
        if bestB==bestA:
            x=random.choice([0,1])
            if x==0:
                for SNP in readSNPs[bestA]:
                    readsA.add(bestA)
                    pos=SNP.split("=")[0]
                    base=SNP.split("=")[1]
                    clusterA.setdefault(pos,{})
                    clusterA[pos].setdefault(base,0)
                    clusterA[pos][base]+=1
                alreadyUsed.add(bestA)
            elif x==1:
                for SNP in readSNPs[bestB]:
                    readsB.add(bestB)
                    pos=SNP.split("=")[0]
                    base=SNP.split("=")[1]
                    clusterB.setdefault(pos,{})
                    clusterB[pos].setdefault(base,0)
                    clusterB[pos][base]+=1
                alreadyUsed.add(bestB)
        else:
            if bestA is not None:
                readsA.add(bestA)
                for SNP in readSNPs[bestA]:
                    pos=SNP.split("=")[0]
                    base=SNP.split("=")[1]
                    clusterA.setdefault(pos,{})
                    clusterA[pos].setdefault(base,0)
                    clusterA[pos][base]+=1
                alreadyUsed.add(bestA)
            if bestB is not None:
                readsB.add(bestB)
                for SNP in readSNPs[bestB]:
                    pos=SNP.split("=")[0]
                    base=SNP.split("=")[1]
                    clusterB.setdefault(pos,{})
                    clusterB[pos].setdefault(base,0)
                    clusterB[pos][base]+=1
                alreadyUsed.add(bestB)

    #Do that iteratively until you have two different clusters and nothing moves anymore
    return clusterA,clusterB,readsA,readsB

def getConsensus(cluster):
    consensusSequence=set()
    for SNP,bases in cluster.items():
        maxBase=None
        maxCov=0
        for base,cov in bases.items():
            if cov>maxCov:
                maxCov=cov
                maxBase=base
        consensusSequence.add(maxBase)
    return consensusSequence

def getRegionCoverage(cluster,allowedPositions):
    totalCoverage=0  #As a %age in the end, but weighted by coverage for each pos

    for SNP,coverage in cluster.items():
        pos=SNP.split("=")[0]
        if pos in allowedPositions:
            totalCoverage+=coverage

    return totalCoverage/len(allowedPositions)

def getRegionDiscordance(cluster,allowedPositions):
    discordanceScore=0  #As a %age in the end, but weighted by coverage for each pos
    totalScore=0

    discordanceDict={}

    for SNP,coverage in cluster.items():
        pos=SNP.split("=")[0]
        if pos in allowedPositions:
            base=SNP.split("=")[1]
            discordanceDict.setdefault(pos,{})
            discordanceDict[pos].setdefault(base,0)
            discordanceDict[pos][base]+=coverage

    for position,bases in discordanceDict.items():
        total=0
        highest=0
        for base,coverage in bases.items():
            total+=coverage
            if coverage>highest:
                highest=coverage
        baseDiscordance=total-highest
        discordanceScore+=baseDiscordance
        totalScore+=total

    if totalScore!=0:
        return discordanceScore/totalScore
    else:
        return 1000000

#Then basically take the regions that are too highly covered and redistribute their reads to the best cluster
#Either it stays in its own cluster, or it goes in another, just pick the best option each time
#That means identify the read that fucks the most with the system, then remove it, then see where it fits best
#Once a read has been moved, it's "safe" and even if placed back in the cluster gets ignored
#Then in the end either you get rid of a cluster or you make it more likely to be stitched back in.

def makeVariantsFile(stitchedClusters,filePath):
    variantFileText=""
    for cluster,clusterData in stitchedClusters.items():
        posDict={}
        for SNP,coverage in clusterData["SNPs"].items():
            chromosome=SNP.split(":")[0]
            position=SNP.split(":")[1].split("=")[0]
            base=SNP.split("=")[1]
            posDict.setdefault(chromosome+":"+position,[0,None])
            if coverage>posDict[chromosome+":"+position][0]:
                posDict[chromosome+":"+position][0]=coverage
                posDict[chromosome+":"+position][1]=base
        for position,data in posDict.items():
            chromosome=position.split(":")[0]
            pos=position.split(":")[1]
            base=data[1]
            variantFileText+="\t".join([cluster,chromosome,pos,base])+"\n"

    variantFile=open(filePath,"w")
    variantFile.write(variantFileText)
    variantFile.close()

def makeClusterReadFile(stitchedClusters,filePath):
    variantFileText=""
    for cluster,clusterData in stitchedClusters.items():
        for read in clusterData["Reads"]:
            variantFileText+=cluster+"\t"+read+"\n"
    variantFile=open(filePath,"w")
    variantFile.write(variantFileText)
    variantFile.close()



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="nPhase",
    version="1.2.0",
    author="Omar Abou Saada",
    author_email="omaroakheart@gmail.com",
    description="nPhase is a command line ploidy agnostic phasing pipeline and algorithm which phases samples of any ploidy with sequence alignment of long and short read data to a reference sequence.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://https://github.com/nPhasePipeline/nPhase",
    entry_points = {
        'console_scripts': [
            'nphase = bin.nPhasePipeline:main'
        ]
    },
    packages=setuptools.find_packages(),
    install_requires=["plotnine","sortedcontainers"],
    classifiers=[
        "Programming Language :: Python :: 3.8",
        "License :: OSI Approved :: GNU General Public License v3 (GPLv3)",
        "Operating System :: OS Independent",
        "Development Status :: 5 - Production/Stable",
        "Topic :: Scientific/Engineering :: Bio-Informatics"
    ],
    python_requires='>=3.8',
)




import operator
import os
from itertools import combinations
import sys
import sortedcontainers
import bin.nPhasePipelineFunctions as nPhaseFunctions
import multiprocessing


def identity(clusterAID,clusterBID,commonPositions):
    allBaseDict={}
    for position in commonPositions:
        baseA=clusterAID[position]
        baseB=clusterBID[position]
        basePossibilities=set().union(baseA["demo"].keys(),baseB["demo"].keys())
        baseNew={"demo":{},"stats":{"N":baseA["stats"]["N"]+baseB["stats"]["N"],"total":baseA["stats"]["total"]+baseB["stats"]["total"]}}
        for base in basePossibilities:
            if baseNew["stats"]["total"]>0:
                baseNew["demo"][base]=(baseA["demo"].get(base,0)*baseA["stats"]["total"]+baseB["demo"].get(base,0)*baseB["stats"]["total"])/baseNew["stats"]["total"]
            else:
                baseNew["demo"][base]=0
        allBaseDict[position]=baseNew
    for position in clusterAID.keys():
        if position not in commonPositions:
            allBaseDict[position]=clusterAID[position]
    for position in clusterBID.keys():
        if position not in commonPositions:
            allBaseDict[position]=clusterBID[position]
    return allBaseDict

def identityChangeBool(clusterID,mergedClusterID,commonPositions,maxID): #Penalize both
    overallChange=0
    allSeqNumbers=0
    if len(commonPositions)==0: #I don't like having this here, but necessary to hack the final stitching, to clean up later. Is it still necessary?
        return False
    for position in commonPositions: #Can be parallelized
        demographics=clusterID[position]
        bestDemo=max(demographics["demo"].values())
        allSeqNumbers+=mergedClusterID[position]["stats"]["N"]
        for demo, proportion in demographics["demo"].items():
            if proportion==bestDemo:
                if mergedClusterID[position]["demo"][demo]-proportion<0: #If it's the best, and after the merge the percentage decreased, then we want to consider that a change in terms of the identity of the cluster
                    overallChange+=abs(mergedClusterID[position]["demo"][demo]-proportion)*mergedClusterID[position]["stats"]["N"]
            else:
                if mergedClusterID[position]["demo"][demo]-proportion>0: #If it isn't the best, and after the merge the percentage increased, then we want to consider that a change in terms of the identity of the cluster
                    overallChange+=abs(proportion-mergedClusterID[position]["demo"][demo])*mergedClusterID[position]["stats"]["N"]
    if allSeqNumbers==0: #Disallow the rare case that two clusters have something in common that isn't supported by any contextDepth info
        return True
    if overallChange/allSeqNumbers>maxID:
        return True
    else:
        return False

def identityTheftBool(positionsA,positionsB,clusterAID,clusterBID,maxID): #There are two possible ways to do this, either by discouraging dissidence, or deposition
    #This is the dissidence mode. Deposition mode would take into account only successful actions against established bases (much more aggressive clustering).
    commonPositions=positionsA&positionsB
    mergedClusterID=identity(clusterAID,clusterBID,commonPositions)
    if identityChangeBool(clusterAID,mergedClusterID,commonPositions,maxID) and identityChangeBool(clusterBID,mergedClusterID,commonPositions,maxID):
        return True
    else:
        return False

def consensus(clusterID):
    consensus=set()
    for position in clusterID.keys():
        highestPct=max(clusterID[position]["demo"].values())
        for base, percentage in clusterID[position]["demo"].items():
            if percentage==highestPct: #To ensure that if there's a 50/50 split, they both get a chance.
                consensus.add(position+"="+base)
    return frozenset(consensus)

def consensusInitializer(sequenceList,contextDepths):
    basePositions=set()
    for sequence in sequenceList:
        basePositions.update([x.split("=")[0] for x in sequence])
    baseDepths={}
    for basePos in basePositions:
        baseDepths[basePos]={}
    for sequence in sequenceList: #Each sequence votes once, therefore the same context is present multiple times in the same cluster results in a stronger vote, hopefully enough to overwhelm chimera but too little to amplify errors
        for base in sequence:
            basePos,baseNt=base.split("=")
            for context, contextCount in contextDepths[base].items():
                if context.issubset(sequence):
                    baseDepths[basePos][baseNt]=baseDepths[basePos].setdefault(baseNt,0)+contextCount
    consensus=set()
    for base, baseDepth in baseDepths.items():
        best=max(baseDepth.values(),default=0)
        for baseNt, depth in baseDepth.items():
            if depth==best:
                consensus.add(base+"="+baseNt)
    return frozenset(consensus)

def nBestPairs(sequences,similarityIndex,minSim,maxID,bannedClusterNames,nPairs):
    indx=0
    indexes=[]
    nBest = []
    bannedOverlaps = set()
    for combination in reversed(similarityIndex):
        indx-=1
        nameA=combination[2]
        nameB=combination[3]
        if nameA in bannedClusterNames or nameB in bannedClusterNames:
            indexes.append(indx)
        else:
            if combination[2] not in bannedOverlaps and combination[3] not in bannedOverlaps:
                cacheA=sequences[combination[2]]
                cacheB=sequences[combination[3]]
                similarity=combination[0]
                if similarity>=minSim:
                    if not identityTheftBool(cacheA["positions"],cacheB["positions"],cacheA["clusterID"],cacheB["clusterID"],maxID):
                        nBest.append([combination[2], combination[3]])
                        bannedOverlaps.update(cacheA["overlaps"], cacheB["overlaps"])
                        if len(nBest)==nPairs:
                            return nBest, indexes
                    else:
                        indexes.append(indx)
                elif similarity<minSim:
                    pass
    return nBest,indexes

def getSimilarity(consensusA,consensusB,commonPos,minOvl,minSim):
    if len(consensusA)<len(consensusB):
        setA=consensusA
        setB=consensusB
    else:
        setA=consensusB
        setB=consensusA
    if len(commonPos)/len(setA)>=minOvl or len(commonPos)>100:
        commonSet=setA&setB
        localSimilarity=len(commonSet)/len(commonPos)
        if localSimilarity<minSim:
            localSimilarity=0
    else:
        localSimilarity=0
    return localSimilarity

def initializeCache(readIDict,name,contextDepths):
    cache={}
    i=0
    for cluster in readIDict.keys():
        cluster=[cluster]
        seqNames=set()
        seqCluster=set()
        for readName in cluster:
            sequence=frozenset([k+"="+next(iter(v["demo"])) for k,v in readIDict[readName].items()])
            seqNames.add(readName)
            seqCluster.add(sequence)
        consensusSeq=consensusInitializer(seqCluster,contextDepths)
        SNPPositions=set([SNP.split("=")[0] for SNP in consensusSeq])
        if len(seqNames)==1:
            clusterID=readIDict[list(seqNames)[0]]
        else:
            print("This broke")
            exit(1)
        cache[name+str(i)]={"names":seqNames,"cluster":seqCluster,"clusterID":clusterID,"consensus":consensusSeq,"positions":SNPPositions,"similarities":{},"overlaps":set()}
        i+=1
    return cache

def generateSimilarityIndex(cache):
    similarityIndex=sortedcontainers.SortedList()
    for simCluster in cache.keys():
        for combination, score in cache[simCluster]["similarities"].items():
            similarPositions=cache[simCluster]["positions"]&cache[combination]["positions"]
            similarityIndex.add([score,len(similarPositions),simCluster,combination])
    return similarityIndex

def cacheFiller(queue,outQueue,minOvl,minSim):
    localCache={}
    while True:
        combinationBatch=queue.get()
        if combinationBatch == "end":
            outQueue.put(localCache)
            return
        for combination in combinationBatch:
            localCache.setdefault(combination[0], {"overlaps": set(), "similarities": {}})
            localCache.setdefault(combination[1], {"overlaps": set(), "similarities": {}})
            cacheKey = combination[0]
            otherKey = combination[1]
            cacheA = cachedSimilarities[cacheKey]
            cacheB = cachedSimilarities[otherKey]
            commonPos = cacheA["positions"] & cacheB["positions"]
            if len(commonPos) > 0:
                localCache[cacheKey]["overlaps"].add(otherKey)
                localCache[otherKey]["overlaps"].add(cacheKey)
                similarity = getSimilarity(cacheA["consensus"], cacheB["consensus"], commonPos, minOvl, minSim)
                if similarity >= minSim:
                    if len(cacheA["consensus"]) < len(cacheB["consensus"]):
                        localCache[cacheKey]["similarities"][otherKey] = similarity
                    else:
                        localCache[otherKey]["similarities"][cacheKey] = similarity

def startCacheFillers(queue, outQueue, cachedSimilarityQueue, cachedSimilarities, minOvl, minSim, threadNumber):
    allProcesses = list()
    for thread in range(0, threadNumber-1):
        cacheFiller_p = multiprocessing.Process(target=cacheFiller, args=((queue),(outQueue),(minOvl),(minSim),))
        cacheFiller_p.daemon = True
        cacheFiller_p.start()
        allProcesses.append(cacheFiller_p)
    cacheMerger_p=multiprocessing.Process(target=cacheMerger, args=((outQueue),(cachedSimilarityQueue),(cachedSimilarities),))
    cacheMerger_p.daemon = True
    cacheMerger_p.start()
    allProcesses.append(cacheMerger_p)

    return allProcesses

def cacheMerger(outQueue,cachedSimilarityQueue,cachedSimilarities):
    while True:
        localDict = outQueue.get()
        if localDict=="end":
            cachedSimilarityQueue.put(cachedSimilarities)
            return
        for cacheName in localDict.keys():
            cachedSimilarities[cacheName]["overlaps"].update(localDict[cacheName]["overlaps"])
            for otherCache, similarity in localDict[cacheName]["similarities"].items():
                cachedSimilarities[cacheName]["similarities"][otherCache] = similarity

def clusterInterlockChimera(minOvl,readIDict,maxID,minLen,minSim,contextDepths,nPairs,threadNumber):
    prevLen=0
    allReadIDict={}
    allReadIDict.update(readIDict)
    #Initializing cachedCluster
    print("Initializing cachedCluster")

    global cachedSimilarities
    cachedSimilarities=initializeCache(readIDict,"cluster",contextDepths)

    #Filling cachedCluster with similarity info
    print("Filling cachedCluster with similarity information")

    batchSize=50000

    #Create a queue for all the combinations
    combinationQueue = multiprocessing.Queue(maxsize=3*threadNumber)
    localSimilarityDictQueue = multiprocessing.Queue()
    cachedSimilarityQueue=multiprocessing.Queue()

    #Start the cache filling processes up before filling up the queue
    allProcesses=startCacheFillers(combinationQueue, localSimilarityDictQueue, cachedSimilarityQueue, cachedSimilarities, minOvl, minSim, threadNumber)
    allCacheFillers=allProcesses[0:-1]
    cacheMergerProcess=allProcesses[-1]


    #Fill up the queue with combinations
    currentBatch=[]
    i=0
    for combination in combinations(cachedSimilarities.keys(),2):
        currentBatch.append(combination)
        if len(currentBatch)>batchSize:
            combinationQueue.put(currentBatch)
            i+=1
            currentBatch=[]

    #Fill the queue with end signals
    for thread in range(0, threadNumber):
        combinationQueue.put("end")

    #Be sure to end all processes
    for cacheFiller in allCacheFillers:
        cacheFiller.join()

    localSimilarityDictQueue.put("end")

    cachedSimilarities=cachedSimilarityQueue.get()

    cacheMergerProcess.join()

    print("Preparing initial similarity index")
    bannedClusterNames=set()
    similarityIndex=generateSimilarityIndex(cachedSimilarities)
    print("Starting clustering loop ("+str(len(cachedSimilarities.keys()))+" sequences)")
    i=0
    #Parallelizing nPhase is very difficult because we cannot create processes with shared access to the cachedSimilarities dict in python.
    #Python can't yet take advantage of Copy On Write optimizations because accessing an object increases its refcount, which counts as a "write".
    #Therefore python is, in practice, Copy On Read, and therefore parallelization requires sending the updated cachedSimilarities dict to processes.
    #Sending such a large dict is prohibitively slow and ruins any improvements we might get from parallelization.
    #Future versions of python may solve this issue, but it seems doubtful. Maybe there's a different data structure I could be using.
    #If you're reading this and have any ideas, send me an email at omaroakheart@gmail.com
    #For more info on python's Copy On Read behavior look up how Instagram circumvented this problem by compiling their own custom version of CPython.
    #
    #Given the impossibility of true parallelization, I decided to at least include an optional heuristic:
    #Allow nPhase to merge more than one cluster per iterative loop.
    #Each iterative loop: Merge together the n (nPairs) non-overlapping best matches.
    #It has a low negative effect on accuracy and low positive effect on runtime in my limited testing.
    #Testing the effect of this parameters on accuracy, contiguity and runtime on large genomes is needed.
    #If you're interested in this, check out my other repo: https://github.com/OmarOakheart/Phasing-Toolkit
    while len(cachedSimilarities.keys())!=prevLen:
        prevLen=len(cachedSimilarities.keys())
        pairs,indexes=nBestPairs(cachedSimilarities,similarityIndex,minSim,maxID,bannedClusterNames,nPairs)
        for indx in reversed(indexes):
            similarityIndex.pop(indx)
        for pair in pairs:
            bannedClusterNames.update(pair)
            if pair!=[]:
                first=pair[0]
                second=pair[1]
                newName="mergedCluster_"+str(i)
                newNames=cachedSimilarities[first]["names"]|cachedSimilarities[second]["names"]
                newCluster=cachedSimilarities[first]["cluster"]|cachedSimilarities[second]["cluster"]
                commonPositions=set([x.split("=")[0] for x in cachedSimilarities[first]["consensus"]])&set([x.split("=")[0] for x in cachedSimilarities[second]["consensus"]])
                newClusterID=identity(cachedSimilarities[first]["clusterID"],cachedSimilarities[second]["clusterID"],commonPositions)
                newOverlaps=cachedSimilarities[first]["overlaps"]|cachedSimilarities[second]["overlaps"]
                newOverlaps.remove(first)
                newOverlaps.remove(second)
                for overlap in newOverlaps:
                    cachedSimilarities[overlap]["overlaps"].add(newName)
                newConsensus=consensus(newClusterID)
                newSNPPositions=set([SNP.split("=")[0] for SNP in newConsensus])
                del cachedSimilarities[first]
                del cachedSimilarities[second]

                newCache = {"names": newNames, "cluster": newCluster, "clusterID": newClusterID, "consensus": newConsensus, "positions": newSNPPositions, "similarities": {}, "overlaps": newOverlaps}
                for cache in newCache["overlaps"]:
                    cacheA = cachedSimilarities[cache]
                    cacheB = newCache
                    if first in cacheA["similarities"]:
                        del cacheA["similarities"][first]
                    if second in cacheA["similarities"]:
                        del cacheA["similarities"][second]
                    if first in cacheA["overlaps"]:
                        cacheA["overlaps"].remove(first)
                    if second in cacheA["overlaps"]:
                        cacheA["overlaps"].remove(second)
                    newCommonPos = cacheA["positions"] & cacheB["positions"]
                    if len(newCommonPos) > minOvl:
                        similarity = getSimilarity(cacheA["consensus"], cacheB["consensus"], newCommonPos, minOvl, minSim)
                    else:
                        similarity = 0
                    if similarity >= minSim:
                        if len(cacheA["consensus"]) < len(cacheB["consensus"]):
                            cacheA["similarities"][newName] = similarity
                            similarityIndex.add([similarity, len(newCommonPos), cache, newName])
                        else:
                            cacheB["similarities"][cache] = similarity
                            similarityIndex.add([similarity, len(newCommonPos), newName, cache])

                cachedSimilarities[newName]=newCache
                i+=1

    #This is where you get the results output, needs another shell script or a python script to do the rest
    actualClusters={}
    resultClusters={}
    for cacheName, cacheInfo in cachedSimilarities.items():
        if len(cacheInfo["names"])>minLen: #Only allow clusters with at least this many reads
            actualClusters[cacheName]=cacheInfo["consensus"]
            resultClusters[cacheName]=cacheInfo
    if len(actualClusters)>0:
        print(len(actualClusters),"clusters")
    else:
        print("No clusters.")
    return actualClusters,resultClusters

#############################
#SPLIT READ CODE STARTS HERE#
#############################

#Using the base cluster info to determine the N SNPs at the edges of the formed clusters

def getSNPPositions(cluster):
    SNPPositions=[]
    for SNP in cluster:
        SNPPositions.append(int(SNP.split("=")[0].split(":")[1]))
    SNPPositions.sort()
    contig=SNP.split(":")[0]
    sortedSNPPositions=[contig+":"+str(SNP) for SNP in SNPPositions]
    return sortedSNPPositions

def getBaseClusterEdges(N,clusters):
    baseClusterEdges=set()
    for cluster in clusters:
        sortedSNPPositions=getSNPPositions(cluster)
        startEdge=frozenset(sortedSNPPositions[0:min(N,len(cluster))])
        endEdge=frozenset(sortedSNPPositions[max(len(cluster)-N,0):len(cluster)])
        baseClusterEdges.add(startEdge)
        baseClusterEdges.add(endEdge)
    return baseClusterEdges

def mergeClusterEdges(clusterEdges):
    newClusterEdges=set()
    newClusterEdges.update(clusterEdges)
    change=True
    while change==True:
        change=False
        for combination in combinations(newClusterEdges,2):
            clusterEdgeA=combination[0]
            clusterEdgeB=combination[1]
            if clusterEdgeA&clusterEdgeB!=set():
                change=True
                if clusterEdgeA in newClusterEdges:
                    newClusterEdges.remove(clusterEdgeA)
                if clusterEdgeB in newClusterEdges:
                    newClusterEdges.remove(clusterEdgeB)
                newClusterEdges.add(frozenset(clusterEdgeA|clusterEdgeB))
                break
    return newClusterEdges

def getSplitReadPositions(chimericReadIDict):
    splitReadPositions=set()
    for splitRead,positions in chimericReadIDict.items():
        splitReadPositions.add(frozenset(positions))
    return splitReadPositions

def getSplitEdgeOverlapProfiles(splitReads,mergedClusterEdges):
    splitReadProfiles=set()
    for splitReadPositions in splitReads:
        splitReadProfile=set()
        for edge in mergedClusterEdges:
            if len(edge&splitReadPositions)>0:
                splitReadProfile.add(edge)
        splitReadProfiles.add(frozenset(splitReadProfile))
    return splitReadProfiles

def getSplitReadProfile(splitReadPositions,mergedClusterEdges):
    splitReadProfile=set()
    for edge in mergedClusterEdges:
        if len(edge&splitReadPositions)>0:
            splitReadProfile.add(edge)
    return splitReadProfile

def flattenProfile(splitReadProfile):
    flattenedProfile=set()
    for edge in splitReadProfile:
        flattenedProfile.update(edge)
    return flattenedProfile

def keepUsefulSplitReadsChr(usefulSplitReadProfiles,chimericReadIDict,mergedClusterEdges):
    usefulSplitReads={}
    for splitRead,splitReadInfo in chimericReadIDict.items():
        splitReadProfile=getSplitReadProfile(set(splitReadInfo.keys()),mergedClusterEdges)
        if splitReadProfile in usefulSplitReadProfiles:
            splitReadPositions=flattenProfile(splitReadProfile)
            newSplitReadInfo={}
            for position,info in splitReadInfo.items():
                if position in splitReadPositions:
                    newSplitReadInfo[position]=info
            chrms=set()
            for position in newSplitReadInfo:
                chrms.add(position.split(":")[0])
            chrmSNPs={}
            for chrm in chrms:
                chrmSNPs[chrm]={}
            for position, info in newSplitReadInfo.items():
                chrmSNPs[position.split(":")[0]][position]=info
            for chrm, posInfo in chrmSNPs.items():
                usefulSplitReads[splitRead]=posInfo
    return usefulSplitReads

def nPhase(longReadSNPAssignments,strainName,contextDepthsFilePath,outFolder,mainFolder,referenceFilePath,minSim,minOvl,minLen,maxID,nPairs,threadNumber):
    #Loading files

    print("Loading reads.")

    SNPAssignmentFile=open(longReadSNPAssignments,"r")

    SNPAssignments={}

    for line in SNPAssignmentFile:
        line=line.strip("\n").split("\t")
        SNPAssignments[line[0]]=frozenset(line[1:])

    SNPAssignmentFile.close()

    print("Reads loaded.")

    #Checking if there are any split reads:
    for v in SNPAssignments.values():
        if len(set(SNP.split(":")[0] for SNP in v))>1:
            print("THERE ARE SPLIT READS IN:",v)
            break

    #######################
    #We need the context depth info:

    print("Loading context depth.")

    contextDepths={}

    contextDepthsLines=[]

    contextDepthsFile=open(contextDepthsFilePath,"r")
    for line in contextDepthsFile:
        line=line.strip("\n").split("\t")
        contextDepthsLines.append(line)

    contextDepthsFile.close()

    for line in contextDepthsLines:
        contextDepths[line[0]]={}
        for context, count in zip(line[1::2],line[2::2]):
            contextDepths[line[0]][frozenset(context.split(" "))]=int(count)

    print("Context depth loaded")

    #This dict will keep track of split reads that we're about to merge
    splitTrackingDict={}

    #Getting chimeric read names
    chimericNames=set()

    for readName, sequence in SNPAssignments.items():
        if len(readName.split("_VCSTTP_"))>1:
            baseName=readName.split("_VCSTTP_")[0]
            chimericNames.add(baseName)
            splitTrackingDict[baseName]={}

    print("Split reads identified.")

    #Applying the context depth info to all of the reads
    chimericReadIDict={}
    for readName, sequence in SNPAssignments.items():
        if len(readName.split("_VCSTTP_"))>1 or readName in chimericNames:
            if len(readName.split("_VCSTTP_"))>1:
                baseName=readName.split("_VCSTTP_")[0]
            if readName in chimericNames:
                baseName=readName
            allBaseDict={}
            if len(sequence)>=10:
                sequence=set(sequence)
                positions=[]
                for base in sequence:
                    chr=base.split(":")[0]
                    if chr not in splitTrackingDict[baseName].keys():
                        splitTrackingDict[baseName][chr]=set()
                    splitTrackingDict[baseName][chr].add(readName)
                    base=base.split("=")
                    positions.append(base[0])
                    allBaseDict.setdefault(base[0],{"demo":{base[1]:0},"stats":{"total":0,"N":0}}).update({"demo":{base[1]:0}})
                    for context, contextCount in contextDepths[base[0]+"="+base[1]].items():
                        if context.issubset(sequence):
                            allBaseDict[base[0]]["demo"][base[1]]+=contextCount
                            allBaseDict[base[0]]["stats"]["total"]+=contextCount
                            allBaseDict[base[0]]["stats"]["N"]+=1
                for basePos, bases in allBaseDict.items():
                    for base in allBaseDict[basePos]["demo"].keys():
                        allBaseDict[basePos]["demo"][base]=allBaseDict[basePos]["demo"][base]/allBaseDict[basePos]["stats"]["total"]
                chimericReadIDict.setdefault("chimeric-"+baseName,{}).update(allBaseDict)

    print("Split reads processed.")

    #Getting all reads
    readIDict={}
    for readName, sequence in SNPAssignments.items():
        if len(sequence)>=10:
            allBaseDict={}
            for base in sequence:
                base=base.split("=")
                allBaseDict.setdefault(base[0],{"demo":{base[1]:0},"stats":{"total":0,"N":0}}).update({"demo":{base[1]:0}})
                for context, contextCount in contextDepths[base[0]+"="+base[1]].items():
                    if context.issubset(sequence):
                        allBaseDict[base[0]]["demo"][base[1]]+=contextCount
                        allBaseDict[base[0]]["stats"]["total"]+=contextCount
                        allBaseDict[base[0]]["stats"]["N"]+=1
            for basePos, bases in allBaseDict.items():
                for base in allBaseDict[basePos]["demo"].keys():
                    allBaseDict[basePos]["demo"][base]=allBaseDict[basePos]["demo"][base]/allBaseDict[basePos]["stats"]["total"]
            readIDict[readName]=allBaseDict

    print("All reads processed.")

    baseClusters,fullClusters=clusterInterlockChimera(minOvl,readIDict,maxID,minLen,minSim,contextDepths,nPairs,threadNumber)
    baseClusters=baseClusters.values()

    baseClusterEdges=getBaseClusterEdges(100,baseClusters)

    mergedClusterEdges=mergeClusterEdges(baseClusterEdges)

    splitReadPositions=getSplitReadPositions(chimericReadIDict)
    splitReadProfiles=getSplitEdgeOverlapProfiles(splitReadPositions,mergedClusterEdges)

    usefulSplitReadProfiles=set()
    for profile in splitReadProfiles:
        if len(profile)>=2:
            usefulSplitReadProfiles.add(profile)

    ##############################################################################################################
    #TODO: Test effect of this additional cluster edge merge on accuracy?

    #mergedSplitReadProfiles=mergeClusterEdges(usefulSplitReadProfiles)
    #newChimericDict = keepUsefulSplitReadsChr(mergedSplitReadProfiles, chimericReadIDict, mergedClusterEdges)
    ##############################################################################################################

    newChimericDict=keepUsefulSplitReadsChr(usefulSplitReadProfiles,chimericReadIDict,mergedClusterEdges)

    cleanAllReadIDict={}
    cleanAllReadIDict.update(readIDict)
    cleanAllReadIDict.update(newChimericDict)

    cleanAllClusters,cleanFullClusters=clusterInterlockChimera(minOvl,cleanAllReadIDict,maxID,minLen,minSim,contextDepths,nPairs,threadNumber)

    #############
    #OUTPUT CODE#
    #############

    visDataTextFull=nPhaseFunctions.giveMeFullData(cleanAllClusters)
    visDataFilePath=os.path.join(outFolder,strainName+"_"+str(minOvl)+"_"+str(minSim)+"_"+str(maxID)+"_"+str(minLen)+"_phasedDataFull.tsv")
    visDataFile=open(visDataFilePath,"w")
    visDataFile.write(visDataTextFull)
    visDataFile.close()

    #Create sets of fastQ read names, to be turned into FASTQ files by another script

    clusterReadText=""

    for clusterName, clusterInfo in cleanFullClusters.items():
        chrms=set()
        for SNP in clusterInfo["consensus"]:
            chrms.add(SNP.split(":")[0])
        if len(chrms)>2:
            print("MULTI-CHROMOSOME CLUSTER ERROR",chrms)
            exit(1)
        if len(chrms)>0:
            for readName in clusterInfo["names"]:
                readName=readName.replace("chimeric-","")
                if readName in splitTrackingDict:
                    splitReads=splitTrackingDict[readName][list(chrms)[0]]
                    for splitRead in splitReads:
                        clusterReadText+="\t".join([clusterName,splitRead])+"\n"
                else:
                    clusterReadText+="\t".join([clusterName,readName])+"\n"

    clusterReadFilePath=os.path.join(outFolder,strainName+"_"+str(minOvl)+"_"+str(minSim)+"_"+str(maxID)+"_"+str(minLen)+"_clusterReadNames.tsv")
    clusterReadFile=open(clusterReadFilePath,"w")
    clusterReadFile.write(clusterReadText)
    clusterReadFile.close()

    #A tab separated file that shows you which SNPs were predicted to be in the same haplotigs

    phasedSNPText=""

    for clusterName, clusterInfo in cleanFullClusters.items():
        for SNP in clusterInfo["consensus"]:
            chrom=SNP.split(":")[0]
            pos=SNP.split(":")[1].split("=")[0]
            base=SNP.split("=")[1]
            phasedSNPText+="\t".join([clusterName,chrom,pos,base])+"\n"

    phasedSNPFilePath=os.path.join(outFolder,strainName+"_"+str(minOvl)+"_"+str(minSim)+"_"+str(maxID)+"_"+str(minLen)+"_variants.tsv")
    phasedSNPFile=open(phasedSNPFilePath,"w")
    phasedSNPFile.write(phasedSNPText)
    phasedSNPFile.close()

    #A tab separated file showing the coverage for each haplotig

    readFilePath=os.path.join(mainFolder,"VariantCalls","longReads",strainName+".hetPositions.SNPxLongReads.validated.tsv")
    outPath=os.path.join(outFolder,strainName+"_"+str(minOvl)+"_"+str(minSim)+"_"+str(maxID)+"_"+str(minLen)+"_covVis.tsv")
    windowSize=10000
    minCov=nPhaseFunctions.generateCoverage(clusterReadFilePath,readFilePath,outPath,windowSize)

    outPath=os.path.join(outFolder,strainName+"_"+str(minOvl)+"_"+str(minSim)+"_"+str(maxID)+"_"+str(minLen)+"_discordanceVis.tsv")
    nPhaseFunctions.generateDiscordance(clusterReadFilePath,readFilePath,outPath)

    outPath=os.path.join(outFolder,strainName+"_"+str(minOvl)+"_"+str(minSim)+"_"+str(maxID)+"_"+str(minLen)+"_minCov="+str(minCov)+"_filterVis.tsv")
    nPhaseFunctions.filterCoverage(clusterReadFilePath,readFilePath,outPath,minCov)

    return "Phasing over"


